// bytran -|- spectral calculations for portable devices using the HITRAN database
//
// Developed by Dzianis Pliutau (Denis Plutov) and Kostya Roslyakov
// www.bytran.org | bytran@bytran.org
// Copyright (c) 2016 Dzianis Pliutau
// Distributed under the MIT license
//
// The MIT License (MIT)
// Copyright (c) 2016 Dzianis Pliutau
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation
// files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,
// modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software
// is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
// OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
// LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
// IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.



// QML component to display the data contained in the atmospheric model used in calculations
// Ground altitude (horizontal path values) will be applied to the current value of temperature, total pressure. pathlength, and partial pressures
import QtQuick 2.9
import QtQuick.Controls 2.2
import QtQuick.Layouts 1.3
import QtQuick.Dialogs 1.2
import QtQuick.Window 2.3
import Plot 1.0

Item {

    property var modelNum

    id: atmModelItem
    anchors.fill: parent

    // Updating the current index (to enable plotting when the form is opened for the second or any subsequent time)
    Component.onCompleted: atmListViewModel.plotSpectrum()

    // Used to rotate the menu
    onHeightChanged: {

        // Resizing controls
        if (Screen.primaryOrientation === Qt.PortraitOrientation)
        {
            modelNum = atmListViewModel.modelNum
            atmModelComboBox.model = [ "Tropical", "Midlatitude\nsummer", "Midlatitude\nwinter", "Subarctic\nsummer", "Subarctic\nwinter", "US Standard\n(default)" ]
            atmListViewModel.modelNum = modelNum
        }
        else
        {
            modelNum = atmListViewModel.modelNum
            atmModelComboBox.model = [ "Tropical", "Midlatitude summer", "Midlatitude winter", "Subarctic summer", "Subarctic winter", "US Standard (default)" ]
            atmListViewModel.modelNum = modelNum
        }
    }

    // Rectangle to draw the form border
    Rectangle {
        anchors.fill: parent
        border.width: 1
        border.color: "black"
        color: Qt.rgba(0, 0, 0, 0.6)

        // This is introduced to disable items located on the background
        MouseArea {
            anchors.fill: parent
            onClicked: { }
        }
    }

    // Rectangle to draw the form border
    Rectangle {
        id: borderRect
        anchors.fill: atmModelColumnLayout
        anchors.margins: coreObj.defaultLabelHeight * (-0.85)
        color: "white"
        border.width: 1
        border.color: "gray"
        radius: coreObj.defaultLabelHeight * 0.5
    }

    // Rectangle to draw the form border around the control elements
    Rectangle {
        id: controlElementsBorderRect
        anchors.fill: controlElementsRowLayout
        anchors.margins: coreObj.defaultLabelHeight * (-0.3)
        anchors.leftMargin: coreObj.defaultLabelHeight * (-0.3)
        color: "white"
        border.width: 1
        border.color: "gray"
        radius: coreObj.defaultLabelHeight * 0.5
    }

    // Column layout to hold all the plot and other interface items
    ColumnLayout {

        // At rotation the buttons are not positioned exactly the way they should be (fix this)
        id: atmModelColumnLayout
        anchors.top: controlElementsBorderRect.bottom
        anchors.bottom: atmModelItem.bottom
        anchors.left: atmModelItem.left
        anchors.right: atmModelItem.right
        anchors.margins: coreObj.defaultLabelHeight * (0.935)
        Layout.fillWidth: true
        spacing: 5
        transformOrigin: Item.Center

        CustomLabel {
            Layout.fillWidth: true
            horizontalAlignment: Text.AlignHCenter
            id: commentLabel
            text: qsTr("Currently only ground altitude values are used.")
        }

        // Rectangle with the plot of the atmospehric model
        Rectangle {
            id: plotRect
            anchors.right: atmModelColumnLayout.right
            anchors.top: commentLabel.bottom
            anchors.bottom: atmModelColumnLayout.bottom
            anchors.left: atmModelColumnLayout.left
            Layout.fillHeight: true
            color: "white"
            border.width: 1
            border.color: "white"
            radius: 3

            Plot {
                id: atmModel1Plot
                anchors.fill: parent
                spectrum: atmListViewModel.spectrum  // Updating the spectrum when the atmospehric model selected has been changed
                minX: atmListViewModel.minX
                maxX: atmListViewModel.maxX
                minY: atmListViewModel.minY
                maxY: atmListViewModel.maxY
                xMinPadding: commentLabel.height * 2.6
                xMaxPadding: atmListView.width * 1.05
                yMinPadding: commentLabel.height * 0.9
                yMaxPadding: commentLabel.height * 2.4
                yAxisTitle: 'Altitude, km'
                xAxisTitle: atmListViewModel.xAxisTitle
                lineColor: atmListViewModel.lineColor
                lineThickness: 2
                numMajorXTicks: 4
                numMajorYTicks: 4
                numMinorXTicks: 4
                numMinorYTicks: 4
                plotInverted: true
                currSpecIndex: atmListViewModel.currSpecIndex
                fontPointSize: ((Qt.platform.os === 'winrt') || (Qt.platform.os === 'winphone')) ? coreObj.defaultFontPointSize : coreObj.defaultFontPointSize + 5 * (1 - coreObj.fontScalingRatio)

                // Used to enable zoom of the plot and other functions
                PinchArea {
                    id: plotPinchArea
                    anchors.fill: parent

//                    // Transfering the pinch gesture coordinates to the plot
//                    onPinchStarted: {
//                        atmModel1Plot.touchCenterX = pinch.center.x;
//                        atmModel1Plot.touchCenterY = pinch.center.y;
//                        if (Math.abs(pinch.point1.x - pinch.point2.x) > Math.abs(pinch.point1.y - pinch.point2.y))
//                             { atmModel1Plot.isHorizontalPinch = true;  }
//                        else { atmModel1Plot.isHorizontalPinch = false; }
//                    }
//                    onPinchUpdated: {
//                        atmModel1Plot.scale = pinch.scale; }
                }
            }
        }

        CustomLabel {
            id: widthLabel
            text: "ClONO2"
            visible: false
            horizontalAlignment: Text.AlignRight
        }

        // Flickable is used to add a scroll bar when the number of items is larger than the screen can fit
        ListView {
            id: atmListView
            anchors.top: commentLabel.bottom
            anchors.topMargin: 5
            anchors.bottomMargin: coreObj.defaultLabelHeight * 2.3
            anchors.right: atmModelColumnLayout.right
            anchors.bottom: atmModelColumnLayout.bottom
            implicitWidth: widthLabel.width * 1.35
            model: atmListViewModel
            spacing: 0
            currentIndex: atmListViewModel.selectedItem
            onCurrentIndexChanged: atmListViewModel.selectedItem = currentIndex
            clip: true

            delegate: RowLayout {
                id: moleculeDelegate
                implicitHeight:  molNameTextLabel.height * 1.4

                CustomSmallerFontText {
                    id: molNameTextLabel
                    text: model.itemName + "                    "
                    horizontalAlignment: Text.AlignRight
                    verticalAlignment: Text.AlignVCenter
                }

                // Mouse area to capture the selection event for a particular item in a list
                MouseArea {
                    anchors.fill: parent
                    onClicked: {
                        atmListView.currentIndex = index;
                    }
                }
            }

            highlight: Rectangle { color: '#D8D8D8' }
        }
    }

    CustomText {
        id: widthText        
        visible: false
        enabled: false
        text: "US Standard (default)"
    }

    RowLayout {
        id: controlElementsRowLayout
        anchors.right: atmModelItem.right
        anchors.top: atmModelItem.top
        anchors.rightMargin: coreObj.defaultLabelHeight * (0.4)
        anchors.topMargin: coreObj.defaultLabelHeight * (0.4)

        // Button to close the form
        CustomButton {
            id: applyValuesButton
            implicitWidth: 1.5 * exitButton.height
            text: qsTr("Apply")

            // Enabling the closing of the form when the exit button is pressed
            onClicked: {
//                if ((Qt.platform.os == 'ios') || (Qt.platform.os == 'windows'))
//                {
                    var component = Qt.createComponent("AtmosphericModelApply.qml")
                    if (component.status === Component.Ready)
                    {
                        atmModelApplyWindow = component.createObject(atmModelItem)
                    }
                    atmModelApplyWindow.show()
//                }
//                else
//                {
//                    applyDialog.show();
//                }
            }
        }

        // ComboBox to select the pathlength units
        CustomComboBox {
            id: atmModelComboBox
            Layout.fillWidth: true
            implicitHeight: exitButton.height
            model: [ "Tropical", "Midlatitude\nsummer", "Midlatitude\nwinter", "Subarctic\nsummer", "Subarctic\nwinter", "US Standard\n(default)" ]
            currentIndex: atmListViewModel.modelNum
            onCurrentIndexChanged: { atmListViewModel.modelNum = currentIndex }
            implicitWidth: atmModelItem.width - applyValuesButton.width - exitButton.width - coreObj.defaultLabelHeight * (0.8) - 2 * controlElementsRowLayout.spacing
        }

        // Button to close the form
        CustomButton {
            id: exitButton            
            implicitWidth: 1.5 * exitButton.height
            text: qsTr("X")

            // Enabling the closing of the form when the exit button is pressed
            onClicked: atmModelItem.destroy();
        }
    }

    // Dialog to exit an applicaion
    MessageDialog {
        id: applyDialog
        icon: StandardIcon.Question
        standardButtons: StandardButton.Yes | StandardButton.No

        function show() {
            applyDialog.text = "The values in the selected atmospheric model at ground altitude (total pressure, temperature, and partial pressures) will be overwritten into the horizontal path calculation variables. Would you like to continue?";
            applyDialog.open()
        }
        onYes: {
            atmListViewModel.assignNewValues();  // Reading in the atmospehric model values for the horizontal path for the current atmospheric model
            molListViewModel.updatePartPress();  // Need to also update the values displayed in the molecules list (partial pressures)
        }
        onNo: { applyDialog.close(); }
    }
}
