// bytran -|- spectral calculations for portable devices using the HITRAN database
//
// Developed by Dzianis Pliutau (Denis Plutov) and Kostya Roslyakov
// www.bytran.org | bytran@bytran.org
// Copyright (c) 2016 Dzianis Pliutau
// Distributed under the MIT license
//
// The MIT License (MIT)
// Copyright (c) 2016 Dzianis Pliutau
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation
// files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,
// modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software
// is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
// OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
// LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
// IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.



// QML component with the configuration parameters for the HITRAN database usage
import QtQuick 2.9
import QtQuick.Controls 2.2
import QtQuick.Layouts 1.3

Item {

    id: hitranConfigItem
    anchors.fill: parent

    // Rectangle to shade the background
    Rectangle {
        anchors.fill: parent
        border.width: 1
        border.color: 'black'
        color: Qt.rgba(0, 0, 0, 0.6)

        // This is introduced to disable items located on the background
        MouseArea {
            anchors.fill: parent
            onClicked: { }
        }
    }

    // Rectangle to draw the form border
    Rectangle {
        anchors.fill: mainHitranConfigColumnLayout
        anchors.margins: coreObj.defaultLabelHeight * (-0.85)
        color: "white"
        border.width: 1
        border.color: 'gray'
        radius: coreObj.defaultLabelHeight * 0.5
    }

    // Exit button
    CustomButton {
        id: exitButton
        anchors.top: hitranConfigItem.top
        anchors.right: hitranConfigItem.right
        anchors.margins: coreObj.defaultLabelHeight * (0.1)
        implicitWidth: 1.5 * exitButton.height
        text: qsTr("X")
        onClicked: hitranConfigItem.destroy()
    }

    CustomLabel {
        id: widthLabel
        text: "HITRAN spectral database usage configuration"
        visible: false
        enabled: false
    }

    // Column layout to hold all the interface items
    ColumnLayout {
        // At rotation the buttons are not positioned exactly the way they should be (fix this)
        id: mainHitranConfigColumnLayout
        anchors.centerIn: parent
        spacing: 5
        transformOrigin: Item.Center
        Layout.fillWidth: true
        Layout.fillHeight: true
        CustomLabel {
            id: commentLabel
            Layout.fillWidth: true            
            text: qsTr("HITRAN spectral database usage configuration")
        }

        CustomDialogButtonWithBackgroundSmall {
            id: authenticateButton
            Layout.fillWidth: true
            text: qsTr("Log in to HITRANOnline")

            // Enabling the closing of the form when the exit button is pressed
            onClicked: {
                var component = Qt.createComponent("Authenticate.qml")
                if (component.status === Component.Ready)
                {
                    authenticateWindow = component.createObject(hitranConfigItem)
                }
                authenticateWindow.show()
            }
            visible: false // This component is currrently not used
        }

        GridLayout {

            // ComboBox to select the pathlength units
            CustomLabel {
                id: storeFormatLabel
                anchors.right: storeFormatComboBox.left
                anchors.rightMargin: coreObj.defaultLabelHeight * 0.3
                Layout.column: 0
                Layout.row: 1
                horizontalAlignment: Text.AlignRight
                text: qsTr("Store database\ndata as:")
            }
            CustomComboBox {
                id: storeFormatComboBox
                Layout.column: 1
                Layout.row: 1
                implicitWidth: widthLabel.width - storageLocationLabel.width
                model: [ "ASCII (*.par files)", "binary (*.bpar files)", "ASCII and binary"]
                currentIndex: hitranDownloadOptionsQml.hitranStoreFileFormat
                onCurrentIndexChanged: hitranDownloadOptionsQml.hitranStoreFileFormat = currentIndex
            }

            // ComboBox to select the pathlength units
            CustomLabel {
                id: dataFormatToUseLabel
                anchors.right: dataFormatToUseComboBox.left
                anchors.rightMargin: coreObj.defaultLabelHeight * 0.3
                Layout.column: 0
                Layout.row: 2
                horizontalAlignment: Text.AlignRight
                text: qsTr("Files to be used\nin calculations:")
            }
            CustomComboBox {
                id: dataFormatToUseComboBox
                Layout.column: 1
                Layout.row: 2
                implicitWidth: widthLabel.width - storageLocationLabel.width
                model: [ "ASCII (*.par files)", "binary (*.bpar files)" ]
                currentIndex: hitranDownloadOptionsQml.hitranFileFormatToUse
                onCurrentIndexChanged: hitranDownloadOptionsQml.hitranFileFormatToUse = currentIndex
            }


            // ComboBox to select the pathlength units
            CustomLabel {
                id: storageLocationLabel
                anchors.right: storeFormatComboBox.left
                anchors.rightMargin: coreObj.defaultLabelHeight * 0.3
                Layout.column: 0
                Layout.row: 3
                horizontalAlignment: Text.AlignRight
                text: qsTr("Storage location:")
                visible: hitranDownloadOptionsQml.externalStoragePermissionGranted
            }
            RowLayout {
                id: storageLocationRowLayout
                Layout.alignment: Qt.AlignHCenter
                Layout.column: 1
                Layout.row: 3
                visible: hitranDownloadOptionsQml.externalStoragePermissionGranted

                CustomLabel {
                    id: internalStorageAuxLabel
                    anchors.right: storageLocationSwitch.left
                    text: qsTr("Internal")
                    color: 'gray'
                    MouseArea {
                        anchors.fill: parent
                        onClicked: hitranDownloadOptionsQml.useSdCard = false
                    }
                }
                CustomSwitch {
                    id: storageLocationSwitch
                    padding: 0
                    checked: hitranDownloadOptionsQml.useSdCard
                    onCheckedChanged: {
                        hitranDownloadOptionsQml.useSdCard = checked
                    }
                    enabled: (hitranDownloadOptionsQml.sdCardPath === "") ? false : true
                }
                CustomLabel {
                    id: sdCardAuxLabel
                    anchors.left: storageLocationSwitch.right
                    text: qsTr("SD card")
                    color: 'red'
                    MouseArea {
                        anchors.fill: parent
                        onClicked: hitranDownloadOptionsQml.useSdCard = true
                    }
                }
            }
        }

        CustomLabel {
            id: noPermissionAuxLabel
            text: {
                      if (Qt.platform.os === 'android') { qsTr("No external storage permission granted. All\ndata is stored in the hidden application folder.") }
                      else { qsTr("No accessible external storage available. All\ndata is stored in the hidden application folder.") }
            }
            visible: !hitranDownloadOptionsQml.externalStoragePermissionGranted
        }

        CustomLabel {
            id: downloadPathLabel
            Layout.fillWidth: true
            horizontalAlignment: Text.AlignRight
            visible: true
            text: {
                if (hitranDownloadOptionsQml.externalStoragePermissionGranted === true)
                {
                    downloadPathLabel.visible = true;
                    // This path is empty if no SD card is detected
                    if (hitranDownloadOptionsQml.sdCardPath === "") { return "No SD memory card detected.\nFolder: " + hitranDownloadOptionsQml.truncatedStoreToPath; }
                    else { return "Folder: " + hitranDownloadOptionsQml.truncatedStoreToPath; }
                }
                else
                {
                    if ((Qt.platform.os === 'ios') || (Qt.platform.os === 'winrt'))
                    {
                        downloadPathLabel.visible = false;
                    }
                    else
                    {
                        downloadPathLabel.visible = true;
                        return "Folder: " + hitranDownloadOptionsQml.storeToPath;
                    }
                }
            }
        }

        CustomDialogButtonWithBackgroundSmall {
            id: setStorageDirButton
            Layout.fillWidth: true            
            text: qsTr("Change storage directory")
            // To be implemented in the future
            // visible: ((Qt.platform.os === 'ios') || (Qt.platform.os === 'android')) ? false : true
            visible: false

            // Enabling the closing of the form when the exit button is pressed
            onClicked: {
                var component = Qt.createComponent("DirDialog.qml")
                if (component.status === Component.Ready)
                {
                    dirDialogWindow = component.createObject(hitranConfigItem)
                }
                dirDialogWindow.show()
            }
        }
    }
}

