// bytran -|- spectral calculations for portable devices using the HITRAN database
//
// Developed by Dzianis Pliutau (Denis Plutov) and Kostya Roslyakov
// www.bytran.org | bytran@bytran.org
// Copyright (c) 2016 Dzianis Pliutau
// Distributed under the MIT license
//
// The MIT License (MIT)
// Copyright (c) 2016 Dzianis Pliutau
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation
// files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,
// modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software
// is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
// OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
// LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
// IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


// This class is used to determine base folder and file paths for data store and access (i.e. HITRAN database downloads etc.)
#include "basepaths.h"
#include <QStandardPaths>
#include <QDir>
#include <QDebug>

BasePaths::BasePaths(QObject *parent,
                     bool newExternalStorageAccessGranted,
                     QString newSdCardBaseDownloadFolderPath) : QObject(parent)
{
    this->useSdCard = false;  // Variable is indicating if the SD card is to be used instead of the internal storage
    this->externalStorageAccessGranted = newExternalStorageAccessGranted;     // Variable is indicating if the external storage access permission has been granted
    this->sDCardBaseDownloadFolderPath = newSdCardBaseDownloadFolderPath;

    // Reading in the state of the variable telling if the SD card is to be used
    readSdCardUseStateFromFile();

    // Creating data directories according to the current settings
    createDirsIfNoneExist();
}

// Methods to save and load the state of if the SD card is to be used instead of the internal storage
void BasePaths::saveSdCardUseStateToFile()
{
    QFile configFile(getInternalApplicaationFolderPath() + "/hiddenconfig.txt");
    QString auxStr;  // Auxiliary string
    QLocale loc(QLocale::English, QLocale::UnitedStates);   // Object to convert to string representatioon and back


    // Maybe better algorithm could be written so that it is not necessary to separately hande the very first and the very last indices
    if (configFile.open(QIODevice::WriteOnly))
    {
        QTextStream ioStream(&configFile);
        if (this->sDCardBaseDownloadFolderPath == "") { auxStr = loc.toString(false) + "\t" + "Use SD card / Use internal storage\n"; }
                                                 else { auxStr = loc.toString(this->useSdCard) + "\t" + "Use SD card / Use internal storage\n"; }
        ioStream << auxStr;   // SD card usage
    }
}
void BasePaths::readSdCardUseStateFromFile()
{
    QFile configFile(getInternalApplicaationFolderPath() + "/hiddenconfig.txt");
    QLocale loc(QLocale::English, QLocale::UnitedStates);      // Object to convert to string representatioon and back
    QString auxStr;  // Auxiliary string variable


    // Maybe better algorithm could be written so that it is not necessary to separately hande the very first and the very last indices
    if (configFile.open(QIODevice::ReadOnly) && this->sDCardBaseDownloadFolderPath != "")
    {
        QTextStream ioStream(&configFile);
        this->useSdCard = loc.toInt((ioStream.readLine().split("\t"))[0]);    // Use SD card / Use internal storage
    }
}

// Folder with the cumulative HITRAN database file (where the complete HITRAN database file is to be stored), this folder will also contain a sub-folder with the HITRAN downloaded in segments from HITRANOnline
QString BasePaths::getBaseDownloadFolderPath()
{
    if (this->externalStorageAccessGranted)  // Using the publicly available directories on the internal storage or on the SD card
    {
        if ((this->useSdCard) && (this->sDCardBaseDownloadFolderPath != ""))  // Using the SD card storage
        {
            return this->sDCardBaseDownloadFolderPath;
        }
        else  // Using the publicly available internal storage
        {
            return QStandardPaths::writableLocation(QStandardPaths::DownloadLocation) + "/bytran";
        }
    }
    else // Using the hidden application directory
    {
        return QStandardPaths::writableLocation(QStandardPaths::AppDataLocation);
    }
}

// This method will return truncated folder paths (initial path removed up to the folders typically visible to the user)
QString BasePaths::getTruncatedBaseDownloadFolderPath()
{
    QString truncatedStr;  // Truncated string
    QString auxStr;   // Auxiliary string
    int index;        // Index in a string


    if (this->externalStorageAccessGranted)  // Using the publicly available directories on the internal storage or on the SD card
    {
        if ((this->useSdCard) && (this->sDCardBaseDownloadFolderPath != ""))  // Using the SD card storage
        {
            index = this->sDCardBaseDownloadFolderPath.lastIndexOf("Android", -1, Qt::CaseInsensitive);
            truncatedStr = this->sDCardBaseDownloadFolderPath.mid(index);
            return truncatedStr;
        }
        else  // Using the publicly available internal storage
        {
            auxStr = QStandardPaths::writableLocation(QStandardPaths::DownloadLocation) + "/bytran";
            index = auxStr.lastIndexOf("Download", -1, Qt::CaseInsensitive);
            truncatedStr = auxStr.mid(index);
            return truncatedStr;
        }
    }
    else // Using the hidden application directory
    {
        auxStr = QStandardPaths::writableLocation(QStandardPaths::AppDataLocation);
        index = auxStr.lastIndexOf("org", -1, Qt::CaseInsensitive);
        truncatedStr = auxStr.mid(index);
        return truncatedStr;
    }
}

// Folder to store the calculation results (images and text files)
QString BasePaths::getResultsFolderPath()
{
    if (this->externalStorageAccessGranted)  // Using the publicly available directories on the internal storage or on the SD card
    {
        if ((this->useSdCard) && (this->sDCardBaseDownloadFolderPath != "")) { return this->sDCardBaseDownloadFolderPath + "/results"; }
                        else { return QStandardPaths::writableLocation(QStandardPaths::DocumentsLocation) + "/bytran"; }
    }
    else  // A path to a sub-folder in the internal application folder
    {
        return QStandardPaths::writableLocation(QStandardPaths::AppDataLocation) + "/results";
    }
}

// Internal application folder (typically used to store the downloaded TIPS files, the molparam.txt, and the globalid.txt)
QString BasePaths::getInternalApplicaationFolderPath()
{
    return QStandardPaths::writableLocation(QStandardPaths::AppDataLocation);
}

// Folder on the SD card with the cumulative HITRAN database file (where the complete HITRAN database file is to be stored), this folder will also contain a sub-folder with the HITRAN downloaded in segments from HITRANOnline
QString BasePaths::getSdCardBaseDownloadFolderPath() { return this->sDCardBaseDownloadFolderPath; }
void BasePaths::setSdCardBaseDownloadFolderPath(QString newSdCardBaseDownloadFolderPath)
{
    this->sDCardBaseDownloadFolderPath = newSdCardBaseDownloadFolderPath;
}

// Variable is indicating if the internal storage or the SD card should be used to store the downloaded HITRAN database and to store the configuration file
bool BasePaths::getUseSdCard() { return this->useSdCard; }
void BasePaths::setUseSdCard(bool newUseSdCard)
{
    this->useSdCard = newUseSdCard;
}
// Methods related to the variable indicating if the external storage access was granted (Android only)
bool BasePaths::getExternalStorageAccessGranted() { return this->externalStorageAccessGranted; }
void BasePaths::setExternalStorageAccessGranted(bool newExternalStorageAccessGranted)
{
    this->externalStorageAccessGranted = newExternalStorageAccessGranted;
}

// Method is used to create directories if they do not exist
void BasePaths::createDirsIfNoneExist()
{
    QDir dir;  // directory variable

    // Setting the application path and creating a directory if it doesn't exist
    this->applicationPath = QStandardPaths::writableLocation(QStandardPaths::AppDataLocation);
    dir.setPath(this->applicationPath);
    if (!dir.exists())
        dir.mkpath(this->applicationPath);

    if (this->externalStorageAccessGranted == true)
    {
        if ((this->useSdCard == true) && (this->sDCardBaseDownloadFolderPath != ""))
        {
            downloadsPath = this->sDCardBaseDownloadFolderPath;  // This is where the HITRAN database will be downloaded to and the confiuration file will be placed into
            documentsPath = this->sDCardBaseDownloadFolderPath + "/results";  // This is where the results of the calculations will be saved to
        }
        else
        {
            // Under android it is possible to save to the Downloads folder if proper permissions are granted (as an alternative to the application folder)
            downloadsPath = QStandardPaths::writableLocation(QStandardPaths::DownloadLocation) + "/bytran";

            // Getting the documents folder path, preparing the folder to save the data to
            documentsPath = QStandardPaths::writableLocation(QStandardPaths::DocumentsLocation) + "/bytran";  // Probably need to move this closer to where the bytran subfolder is created
        }
    }
    else // External storage permission has not been granted, using the internal application folder to store all data
    {
        // By default the HITRAN downloads will be stored in the internal application folder (only under Android saving under other location is possible)
        downloadsPath = QStandardPaths::writableLocation(QStandardPaths::AppDataLocation);

        // By default the data folder is within the internal application folder path
        documentsPath = QStandardPaths::writableLocation(QStandardPaths::AppDataLocation) + "/results";
    }

    // Creating the Downloads folder if it does not exist
    dir.setPath(downloadsPath);
    if (!dir.exists())
        dir.mkpath(downloadsPath);

    // Creating the Documents folder if it does not exist
    dir.setPath(documentsPath);
    if (!dir.exists())
        dir.mkpath(documentsPath);

}
