// bytran -|- spectral calculations for portable devices using the HITRAN database
//
// Developed by Dzianis Pliutau (Denis Plutov) and Kostya Roslyakov
// www.bytran.org | bytran@bytran.org
// Copyright (c) 2016 Dzianis Pliutau
// Distributed under the MIT license
//
// The MIT License (MIT)
// Copyright (c) 2016 Dzianis Pliutau
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation
// files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,
// modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software
// is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
// OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
// LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
// IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.



// This class is used to download data from the HITRAN online interface. The data includes the
// molparam.txt file, the files with the total internal partition sums and the HITRAN database file.
// The module also includes a subclass to access HITRAN database in a standard format of a disk.
#ifndef HITRAN_H
#define HITRAN_H

#include <QObject>
#include <QString>
#include "hitrandownloadhapi.h"
//#include "hitrandownloadweb.h"
#include "hitrandownloadregionsfinder.h"
#include "hitraninputoutput.h"
#include "filedownloader.h"
#include "tipsfilesdownloader.h"
#include "hitranonerec.h"

class Hitran : public QObject
{
    Q_OBJECT

public:

    explicit Hitran(QObject *parent = nullptr);

    // Getting and setting the minimum and maximum download wavenumbers
    double getStartWn();
    void setStartWn(double newStartWn);
    double getEndWn();
    void setEndWn(double newEndWn);
    // Number of the start record in the cumulative HITRAN data file
    quint64 getStartWnByte();
    void setStartWnByte(quint64 newStartWnByte);
    // Number of the end record in the cumulative HITRAN data file
    quint64 getEndWnByte();
    void setEndWnByte(quint64 newEndWnByte);
    // Number of the starting subrange record
    quint64 getStartSubrangeWnByte();
    void setStartSubrangeWnByte(quint64 newStartSubrangeWnByte);
    // Number of the ending subrange record
    quint64 getEndSubrangeWnByte();
    void setEndSubrangeWnByte(quint64 newEndSubrangeWnByte);
    // Number of records in the subrange to read in one reading
    quint32 getSubrangeRecordsStep();
    void setSubrangeRecordsStep(quint32 newSubrangeRecordsStep);

    // Getting and setting the variable indicating the type of HITRAN database usage to implement
    quint8 getHitranAccessMode();
    void setHitranAccessMode(quint8 newHitranAccessMode);

    // Methods to get and set the host name to download the HITRAN database
    QString getLblDataHostName();
    void setLblDataHostName(QString newLblDataHostName);
    // Method to set the HITRAN hostname, email and password
    //QString getHitranLoginHost();
    //void setHitranLoginHost(QString newHitranLoginHostName);
    // Getting and setting the HITRAN email
    //QString getEmail();
    //void setEmail(QString newEmail);
    // Getting and setting the HITRAN password
    //QString getPassword();
    //void setPassword(QString newPassword);
    QString getHapiToken();
    void setHapiToken(QString newHapiToken);

    // Setting the variable indicating if authentication is to be used in remote data access
    // Currently the only authentication method is using the HITRAN Web-based access, HAPI authentication will be added in the future
    bool getUserAuthen();
    void setUserAuthen(bool newUserAuthen);

    // Method to return a reference to the hitranio class
    HitranInputOutput *getHitranIO();

    // Method to get the object used to connect to HITRANonline over the web
    //HitranDownloadWeb *getHitranWeb();
    // Method to get the object used to connect to HITRANonline using HAPI style access
    HitranDownloadHapi *getHitranHapi();

    // Method to set a reference to the molecules array
    void setMolArray(QVector<HitranOneMol> *newMol);

    // Method to download HITRAN
    void downloadHitranData();
    void downloadHitranData(bool newUseAuthen);
    // There is one more method called getHitranData but it is private

    // Method is to carry out complete HITRAN online database download (segmented files which will be stored separately)
    void downloadCompleteSegmentedHitran();

    // Method to get data to be used in calculations
    QVector<HitranOneRec> *getHitranRecs();

    // Method to initiate the HITRAN download query
    QString formLblDataQuery(double fromWn, double toWn);

    // Method to determine if the complete HITRAN database file is present on the user's device
    int findCumulativeOrSegmentedHitranFiles();

    // Method to get the number of records to be read in from the HITRAN file
    quint64 getTotalNumRecsToProcess();

    // Method to determine the HITRAN database record boundaries for the entire spectral region based on the start and end wavenumbers
    void findHitranRecordBoundaries();

    // Method to get the next records subrange
    void getNextSubrange();

    // Method is used to set a variable used to cancel the calculation
    void setCancelCalculation(bool newCancelCalculation);

signals:

    // Signal used to send notifications to the user of what is going on in the program
    void notify(QString message, double value, bool visible, bool cancelButtonVisible);

    // Signal is emitted after the download of the HITRAN database
    void hitranDownloadComplete();
    // Signal is emited if the HITRAN database download is unsuccessful
    void hitranDownloadUnsuccessful();

    // Signal is emited when the data becomes available (either read in from a file or downloaded)
    void dataAvailable();
    // Emitted when no more data is available
    void noMoreDataAvailable();

    // Singnal is emited to open the ok notification dialog
    void openOkDialog(QString text);

private slots:

    // Slot indicating that authentication has been completed
    void authenticationComplete();

    // Method to send an initial download request
    void sendInitialDownloadRequest();

    // Slot to process the server downloaded data
    void processServerReturnedData(QByteArray *hitranData, bool success);

private:

    // Private method to get HITRAN data over a specified spectral region
    void downloadHitranData(bool newUseAuthen, double startWn, double endWn);

    // Method called to request data from a remote server
    void sendDataRequest(double fromWn, double toWn);
    void sendDownloadEveryTimeDataRequest();

    // Method to create actions (connections to carry out further operations)
    void createActions();

    // Sub-query string (line-by-line)
    QString lblDataHostName;                  // http://www.hitran.org/lbl/ - Web address to request the line-by-line data from
    QString lblSubQueryStrWithAuthen;     // String to contain additional characters for line-by-line query with authentication
    QString lblSubQueryStrWithoutAuthen;  // String to contain additional characters for line-by-line query without authentication

    // HITRAN database wavenumber range related variables
    double startWn;  // Minimum download wavenumber
    double endWn;    // Maximum download wavenumber
    quint64 startWnByte;  // Number of the start record in the cumulative HITRAN data file
    quint64 endWnByte;    // Number of the end record in the cumulative HITRAN data file
    quint64 startSubrangeWnByte;  // Number of the starting subrange record byte
    quint64 endSubrangeWnByte;    // Number of the ending subrange record byte
    quint32 subrangeRecordsStep;  // Number of records (not bytes) in the subrange to read in one reading

    // HITRAN database segmented files related variables
    int startWnFileIndex;  // Starting file index for the segmented HITRAN access
    int endWnFileIndex;    // Ending file index for the segmented HITRAN access
    int startSubrangeWnFileIndex;  // Starting subrange file index for the segmented HITRAN access
    int endSubrangeWnFileIndex;    // Ending subrange file index for the segmented HITRAN access

    // Objects to retrieve HITRAN data from a remote server over Web or using the HAPI style access
    HitranDownloadHapi *hitranHapi;
    //HitranDownloadWeb *hitranWeb;

    // Object to determine the download regions for data
    HitranDownloadRegionsFinder *downloadRegionsFinder;

    // Object for saving of data to the HITRAN file
    HitranInputOutput *hitranio;

    // Variable indicates if authentication needs to be used when accessing the remote server
    bool userAuthen;

    // Data downloaded for each spectral data region
    QStringList downloadedData;

    // Variable indicating the number of the HITRAN downloaded regions (use used to only update the download history file one time)
    int numHitranRegionsDownloaded;

    // Variable indicating if the HITRAN database should be accessed of the disk or downloaded (thd download option is the default)
    quint8 hitranAccessMode;  // 0 - download and save HITRAN file in segments as the need arises and store them locally for later reuse offline (previously unaccessed segments would have to be downloaded);
                              // 1 - use cumulative HITRAN file (download complete HITRAN file once only and later use it offline);
                              // 2 - dowload HITRAN segments from hitran.org every time, no storing of data on the mobile device (not recommended)

    // Array with information about molecules and isotopes
    QVector<HitranOneMol> *mol;

    // Variable set to cancel the calculation
    bool cancelCalculation;

};

#endif // HITRAN_H
