// bytran -|- spectral calculations for portable devices using the HITRAN database
//
// Developed by Dzianis Pliutau (Denis Plutov) and Kostya Roslyakov
// www.bytran.org | bytran@bytran.org
// Copyright (c) 2016 Dzianis Pliutau
// Distributed under the MIT license
//
// The MIT License (MIT)
// Copyright (c) 2016 Dzianis Pliutau
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation
// files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,
// modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software
// is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
// OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
// LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
// IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


// This object is used to download auxiliary HITRAN data (molparam.txt, TIPS files and the Global ids file)
#include "hitranauxiliary.h"
#include <QFileInfo>

HitranAuxiliary::HitranAuxiliary(QObject *parent) : QObject(parent)
{
    // File downloader object to download the TIPS files and the file with the molecular information
    fileDownloader = new FileDownloader();

    // Linking the download object to the corresponding slot for the molparam.txt and the globalid.txt files download
    connect(this->fileDownloader, SIGNAL(dataDownloaded(bool)), this, SLOT(molparamOrGlobalidDownloaded(bool)));

    // Initializing the TIPS downloader
    tipsDownloader = new TipsFilesDownloader();

    // Linking signal slot connections for the TIPS downloads
    connect(this->tipsDownloader, SIGNAL(notify(QString, double, bool, bool)), this, SIGNAL(notify(QString, double, bool, bool)));
    connect(this->tipsDownloader, SIGNAL(tipsFilesDownloaded(bool)), this, SLOT(tipsFilesDownloaded(bool)));
}

// Path to the folder where HITRAN, molparam.txt, and the globalid.txt files are stored
QString HitranAuxiliary::getMolparamFolder() { return this->molparamFolder; }
void HitranAuxiliary::setMolparamFolder(QString newMolparamFolder) { this->molparamFolder = newMolparamFolder; }

// Paths to the online download folders with the molparam and the TIPS files
QString HitranAuxiliary::getMolparamDownloadAddressPath() { return this->molparamDownloadAddressPath; }
void HitranAuxiliary::setMolparamDownloadAddressPath(QString newMolparamDownloadAddressPath) { this->molparamDownloadAddressPath = newMolparamDownloadAddressPath; }
// Paths to the online download folders with the molparam and the TIPS files
QString HitranAuxiliary::getGloabalIdTxtDownloadAddressPath() { return this->globalIdTxtDownloadAddressPath; }
void HitranAuxiliary::setGlobalIdTxtDownloadAddressPath(QString newGlobalIdTxtDownloadAddressPath) { this->globalIdTxtDownloadAddressPath = newGlobalIdTxtDownloadAddressPath; }

// Method to get the TIPS object
TipsFilesDownloader *HitranAuxiliary::getTipsDownloader() { return this->tipsDownloader; }

void HitranAuxiliary::downloadMolparamTxt()
{
    QString fileName;  // File name extracted from the download path
    int index;  // Index within the string to start the download at

    index = this->molparamDownloadAddressPath.lastIndexOf("/");
    fileName = this->molparamDownloadAddressPath.mid(index);
    QFileInfo checkFile(this->molparamFolder + fileName);  //  "/" is not added because it is already there

    if (!checkFile.exists())
    {
        fileDownloader->setDownloadPath(this->molparamDownloadAddressPath);
        fileDownloader->setSavePath(this->molparamFolder + fileName);
        fileDownloader->setSaveToHarddrive(true);
        fileDownloader->download();
    }
    else
    {
        // Downloading TIPS files after the molparam files have been downloaded
        emit molparamTxtDownloadComplete();
    }
}
// Slot is called after the molparam.txt file has been downloaded
void HitranAuxiliary::molparamOrGlobalidDownloaded(bool success)
{
    if (success == false)
    {
        if (fileDownloader->getSavePath().toLower().contains("molparam.txt")) {
            emit openOkDialog("Error downloading the molparam.txt\nfile from HITRAN online (hitran.org)");
        }
        if (fileDownloader->getSavePath().toLower().contains("globalid.txt")) {
            emit openOkDialog("Error downloading the globalid.txt\nfile from the bytran website (bytran.org)");
        }
    }
    else
    {
        if (fileDownloader->getSavePath().toLower().contains("molparam.txt")) {
            // Emiting a signal indicating that the download of the molparam.txt file is complete
            emit molparamTxtDownloadComplete();
        }
        if (fileDownloader->getSavePath().toLower().contains("globalid.txt")) {
            // Emiting a signal indicating that the download of the globalid.txt file is complete
            emit globalIdTxtDownloadComplete();
        }
    }
}

// Method to download files with the total internal partition sums
void HitranAuxiliary::downloadTipsFiles()
{
    tipsDownloader->downloadTipsFiles();
}
// Slot is called once the TIPS files have finished downloading
void HitranAuxiliary::tipsFilesDownloaded(bool success)
{
    if (success == false)
    {
        emit openOkDialog("Error downloading TIPS files\nfrom HITRANonline (hitran.org)");
    }
    else
    {
        // Emitting a signal indicating that auxiliary data download is complete
        emit tipsDownloadComplete();
    }
}

// Method to download the custom file with the isotopes and their corresponding global ID number
void HitranAuxiliary::downloadGlobalIdTxt()
{
    QString fileName;  // File name extracted from the download path
    int index;  // Index within the string to start the download at

    index = this->globalIdTxtDownloadAddressPath.lastIndexOf("/");
    fileName = this->globalIdTxtDownloadAddressPath.mid(index);
    QFileInfo checkFile(this->molparamFolder + fileName);  //  "/" is not added because it is already there

    if (!checkFile.exists())
    {
        fileDownloader->setDownloadPath(this->globalIdTxtDownloadAddressPath);
        fileDownloader->setSavePath(this->molparamFolder + fileName);
        fileDownloader->setSaveToHarddrive(true);
        fileDownloader->download();
    }
    else
    {
        // Downloading TIPS files after the molparam files have been downloaded
        emit globalIdTxtDownloadComplete();
    }
}
