// bytran -|- spectral calculations for portable devices using the HITRAN database
//
// Developed by Dzianis Pliutau (Denis Plutov) and Kostya Roslyakov
// www.bytran.org | bytran@bytran.org
// Copyright (c) 2016 Dzianis Pliutau
// Distributed under the MIT license
//
// The MIT License (MIT)
// Copyright (c) 2016 Dzianis Pliutau
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation
// files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,
// modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software
// is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
// OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
// LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
// IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.



// Data object to specify the HITRAN database download options
#include "hitrandownloadoptionsqml.h"
#include <QMessageBox>

HitranDownloadOptionsQml::HitranDownloadOptionsQml(QObject *parent, BytranCore *inCore) : QObject(parent)
{
    // Passing in a reference to the core object
    core = inCore;

    // Saving the value of the HITRAN database usage parameter
    m_hitranFileUsage = core->getHitran()->getHitranAccessMode();
    m_hitranStoreFileFormat = core->getHitran()->getHitranIO()->getHitranStoreFileFormat();
    m_hitranFileFormatToUse = core->getHitran()->getHitranIO()->getHitranFileFormatToUse();
    connect(this->core, SIGNAL(valuesUpdated()), this, SIGNAL(hitranFileUsageChanged()));
}

quint8 HitranDownloadOptionsQml::hitranFileUsage() const { return core->getHitran()->getHitranAccessMode(); }
void HitranDownloadOptionsQml::setHitranFileUsage(const quint8 hitranFileUsage)
{
    if (hitranFileUsage != m_hitranFileUsage)
    {
        m_hitranFileUsage = hitranFileUsage;
        core->getHitran()->setHitranAccessMode(m_hitranFileUsage);
        emit hitranFileUsageChanged();
    }
}

// Format to store the HITRAN database in (ASCII, binary or both)
quint8 HitranDownloadOptionsQml::hitranStoreFileFormat() const { return this->core->getHitran()->getHitranIO()->getHitranStoreFileFormat(); }
void HitranDownloadOptionsQml::setHitranStoreFileFormat(const quint8 hitranStoreFileFormat)
{
    if (hitranStoreFileFormat != m_hitranStoreFileFormat)
    {
        m_hitranStoreFileFormat = hitranStoreFileFormat;
        core->getHitran()->getHitranIO()->setHitranStoreFileFormat(m_hitranStoreFileFormat);
        emit hitranStoreFileFormatChanged();
    }
}

// Format of the HITRAN database to be used in calculations (ASCII vs. binary)
quint8 HitranDownloadOptionsQml::hitranFileFormatToUse() const { return this->core->getHitran()->getHitranIO()->getHitranFileFormatToUse(); }
void HitranDownloadOptionsQml::setHitranFileFormatToUse(const quint8 hitranFileFormatToUse)
{
    if (hitranFileFormatToUse != m_hitranFileFormatToUse)
    {
        m_hitranFileFormatToUse = hitranFileFormatToUse;
        core->getHitran()->getHitranIO()->setHitranFileFormatToUse(m_hitranFileFormatToUse);
        emit hitranFileFormatToUseChanged();
    }
}

// Boolean variable indicating if the downloaded HITRAN data should be stored on an internal card or an external storage
bool HitranDownloadOptionsQml::useSdCard() const { return core->getBasePaths()->getUseSdCard(); }
void HitranDownloadOptionsQml::setUseSdCard(const bool useSdCard)
{
    if (useSdCard != m_useSdCard)
    {
        m_useSdCard = useSdCard;
        core->getBasePaths()->setUseSdCard(m_useSdCard);
        emit useSdCardChanged();
        emit storeToPathChanged();
        emit truncatedStoreToPathChanged();
        emit sdCardPathChanged();
    }
}

// Variable to return the path to the downloads folder on an SD card
QString HitranDownloadOptionsQml::sdCardPath() const { return core->getBasePaths()->getSdCardBaseDownloadFolderPath(); }

// Path to where the downloaded HITRAN database and the configuration file are to be stored
QString HitranDownloadOptionsQml::truncatedStoreToPath() const { return core->getBasePaths()->getTruncatedBaseDownloadFolderPath(); }
QString HitranDownloadOptionsQml::storeToPath() const { return core->getBasePaths()->getBaseDownloadFolderPath(); }

// Variable indicating if the external storage permission has been granted
bool HitranDownloadOptionsQml::externalStoragePermissionGranted() const { return core->getPermissions()->getPermissionResult(); }

// Method to check if the HITRAN database exists on the server
bool HitranDownloadOptionsQml::cumulativeHitranExists() const
{
    int value;

    value = core->getHitran()->findCumulativeOrSegmentedHitranFiles();
    if ((value == 10) || (value == 11) || (value == 12) || (value == 1))
    {
        return true;
    }
    else
    {
        return false;
    }
}

// Method to download the complete HITRAN database
void HitranDownloadOptionsQml::downloadCompleteHitran() { core->getHitran()->downloadCompleteSegmentedHitran(); }

// Method to create directories if they do not exist
void HitranDownloadOptionsQml::createDirsIfNoneExist()
{
    // Updating the save directories
    core->getBasePaths()->createDirsIfNoneExist();

    // Saving the updated data save directory path to the HitranIO object
    core->getHitran()->getHitranIO()->setBaseDownloadFolderPath(core->getBasePaths()->getBaseDownloadFolderPath());
}
