// bytran -|- spectral calculations for portable devices using the HITRAN database
//
// Developed by Dzianis Pliutau (Denis Plutov) and Kostya Roslyakov
// www.bytran.org | bytran@bytran.org
// Copyright (c) 2016 Dzianis Pliutau
// Distributed under the MIT license
//
// The MIT License (MIT)
// Copyright (c) 2016 Dzianis Pliutau
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation
// files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,
// modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software
// is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
// OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
// LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
// IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.



// Data object to specify the HITRAN database download options
#ifndef HITRANDOWNLOADOPTIONSQML_H
#define HITRANDOWNLOADOPTIONSQML_H

#include <QObject>
#include "bytrancore.h"

class HitranDownloadOptionsQml : public QObject
{
    Q_OBJECT

    Q_PROPERTY(quint8 hitranFileUsage READ hitranFileUsage WRITE setHitranFileUsage NOTIFY hitranFileUsageChanged)
    Q_PROPERTY(bool cumulativeHitranExists READ cumulativeHitranExists NOTIFY cumulativeHitranExistsChanged)
    Q_PROPERTY(quint8 hitranFileFormatToUse READ hitranFileFormatToUse WRITE setHitranFileFormatToUse NOTIFY hitranFileFormatToUseChanged)
    Q_PROPERTY(quint8 hitranStoreFileFormat READ hitranStoreFileFormat WRITE setHitranStoreFileFormat NOTIFY hitranStoreFileFormatChanged)
    Q_PROPERTY(bool useSdCard READ useSdCard WRITE setUseSdCard NOTIFY useSdCardChanged)
    Q_PROPERTY(QString sdCardPath READ sdCardPath NOTIFY sdCardPathChanged)
    Q_PROPERTY(QString truncatedStoreToPath READ truncatedStoreToPath NOTIFY storeToPathChanged)
    Q_PROPERTY(QString storeToPath READ storeToPath NOTIFY storeToPathChanged)
    Q_PROPERTY(bool externalStoragePermissionGranted READ externalStoragePermissionGranted NOTIFY externalStoragePermissionGrantedChanged)

public:

    // Method to download the entire HITRAN database
    Q_INVOKABLE void downloadCompleteHitran();

    // Method to create directories if they do not exist
    Q_INVOKABLE void createDirsIfNoneExist();

    explicit HitranDownloadOptionsQml(QObject *parent = nullptr, BytranCore *inCore = nullptr);

    // Methods to get and set the HITRAN file usage (download vs. local etc.)
    quint8 hitranFileUsage() const;
    void setHitranFileUsage(const quint8 hitranFileUsage);

    // Format to store the HITRAN database in (ASCII, binary or both)
    quint8 hitranStoreFileFormat() const;
    void setHitranStoreFileFormat(const quint8 hitranStoreFileFormat);

    // Format of the HITRAN database to be used in calculations (ASCII vs. binary)
    quint8 hitranFileFormatToUse() const;
    void setHitranFileFormatToUse(const quint8 hitranFileFormatToUse);

    // Boolean variable indicating if the downloaded HITRAN data should be stored on an internal card or an external storage
    bool useSdCard() const;
    void setUseSdCard(const bool useSdCard);

    // Variable to return the path to the downloads folder on an SD card
    QString sdCardPath() const;

    // Path to where the downloaded HITRAN database and the configuration file are to be stored
    QString truncatedStoreToPath() const;
    QString storeToPath() const;

    // Variable indicating if the external storage permission has been granted
    bool externalStoragePermissionGranted() const;

    // Method returns a value indicating if the cumulative HITRAN database file exists
    bool cumulativeHitranExists() const;

signals:

    void hitranFileUsageChanged();
    void cumulativeHitranExistsChanged();
    void hitranStoreFileFormatChanged();
    void hitranFileFormatToUseChanged();
    void useSdCardChanged();
    void sdCardPathChanged();
    void storeToPathChanged();
    void truncatedStoreToPathChanged();
    void externalStoragePermissionGrantedChanged();

public slots:

private:

    // Core object related to the main calculation parameters
    BytranCore *core;

    // Variable indicating if the HITRAN database should be accessed of the disk or downloaded (thd download option is the default)
    quint8 m_hitranFileUsage;  // 0 - download and save HITRAN file in segments as the need arises and store them locally for later reuse offline (previously unaccessed segments would have to be downloaded);
                               // 1 - use cumulative HITRAN file (download complete HITRAN file once only and later use it offline);
                               // 2 - dowload HITRAN segments from hitran.org every time, no storing of data on the mobile device (not recommended)

    bool m_cumulativeHitranExistsChanged;  // true - cumulative HITRAN file exists; false - cumulative HITRAN file does not exist

    // Variable indicating if the downloaded HITRAN data to be stores as ASCII, binary or both
    quint8 m_hitranStoreFileFormat;  // Store downloaded HITRAN files as: 0 - ASCII, 1 - binary, 2 - ASCII and binary
    // Variable indicating if the HITRAN data to be used is ASCII or binary
    quint8 m_hitranFileFormatToUse;  // Use HITRAN saved in one of the following formats: 0 - ASCII, 1 - binary
    bool m_useSdCard;  // Boolean variable indicating if the downloaded HITRAN data should be stored on an internal card or an external storage
    QString m_sdCardPath;  // Path to the SD card
    QString m_storeToPath;   // Path to where the downloaded HITRAN database and the configuration file are to be stored
    QString m_truncatedStoreToPath;  // Path containing only the subfolders visible to the user
    bool m_externalStoragePermissionGranted;  // Variable indicating if the external storage permission has been granted

};

#endif // HITRANDOWNLOADOPTIONSQML_H
