// bytran -|- spectral calculations for portable devices using the HITRAN database
//
// Developed by Dzianis Pliutau (Denis Plutov) and Kostya Roslyakov
// www.bytran.org | bytran@bytran.org
// Copyright (c) 2016 Dzianis Pliutau
// Distributed under the MIT license
//
// The MIT License (MIT)
// Copyright (c) 2016 Dzianis Pliutau
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation
// files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,
// modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software
// is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
// OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
// LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
// IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.



// Object to read/write HITRAN lines within a specified spectral range
#include "hitraninputoutput.h"
#include "const.h"
#include <QLocale>
#include <QFile>
#include <QTextStream>
#include <QApplication>
#include <math.h>
#include <QDir>
#include <QFloat16>
#include <QStandardPaths>
#include <QMessageBox>
#include <QDebug>

HitranInputOutput::HitranInputOutput()
{    
    // Variables related to the HITRAN database storage location (total HITRAN or segmented)
    this->cumulativeHitranFileName = "HITRAN2012.par";   // Name of the file to save the downloaded data to (cumulative HITRAN file name)
    this->segmentedHitranSubFolderName = "HITRAN-downloads";     // Folder to contain the segmented HITRAN database files in the ASCII format and the subfolder for the binary format files
    this->segmentedHitranBinSubFolderName = "BIN";  // Subfolder to contain segmented HITRAN files in a binary format

    // Variable indicating if the downloaded HITRAN data to be stores as ASCII, binary or both
    this->hitranStoreFileFormat = 2;   // Store downloaded HITRAN files as: 0 - ASCII, 1 - binary, 2 - ASCII and binary

    // Variable indicating if the HITRAN data to be used is ASCII or binary
    this->hitranFileFormatToUse = 1;   // Use HITRAN saved in one of the following formats: 0 - ASCII, 1 - binary
}

// HITRAN database local storage location related methods
// Name of the total HITRAN database file (the one which contains all records, old *.par cumulative file style)
QString HitranInputOutput::getCumulativeHitranFileName() { return this->cumulativeHitranFileName; }
void HitranInputOutput::setCumulativeHitranFileName(QString newCumulativeHitranFileName) { this->cumulativeHitranFileName = newCumulativeHitranFileName; }
// Folder with the total HITRAN database file (where the complete HITRAN database file is to be stored)
QString HitranInputOutput::getBaseDownloadFolderPath() { return this->baseDownloadFolderPath; }
void HitranInputOutput::setBaseDownloadFolderPath(QString newBaseDowloadFolderPath)
{
    // Changing the base downloads folder path
    this->baseDownloadFolderPath = newBaseDowloadFolderPath;

    // Create the HITRAN data storage directories if they do not exist
    createHitranDataDirsIfNoneExist();
}
// Method will check if the ASCII and binary HITRAN folders exist and create them if the do not exist
void HitranInputOutput::createHitranDataDirsIfNoneExist()
{
    QDir dir;

    // The creation of these folders should probably be moved into the HitranIO class
    dir.setPath(this->getSegmentedHitranFolderPath());
    if (!dir.exists())
        dir.mkpath(this->getSegmentedHitranFolderPath());  // This may require a permission

    // Creating a HITRAN segmented files sub-directory for HITRAN files in a binary format (*.bpar)
    dir.setPath(this->getSegmentedHitranBinFolderPath());
    if (!dir.exists())
       dir.mkpath(this->getSegmentedHitranBinFolderPath());  // This may require a permission
}
// Method to get and set the segmented HITRAN folder
QString HitranInputOutput::getSegmentedHitranSubFolderName() { return this->segmentedHitranSubFolderName; }
void HitranInputOutput::setSegmentedHitranSubFolderName(QString newSegmentedHitranSubFolderName) { this->segmentedHitranSubFolderName = newSegmentedHitranSubFolderName; }
// Folder to contain the segmented HITRAN database files in binary format
QString HitranInputOutput::getSegmentedHitranBinSubFolderName() { return this->segmentedHitranBinSubFolderName; }
void HitranInputOutput::setSegmentedHitranBinSubFolderName(QString newSegmentedHitranBinSubFolder) { this->segmentedHitranBinSubFolderName = newSegmentedHitranBinSubFolder; }

// Folder to contain the segmented HITRAN database files in ASCII format
QString HitranInputOutput::getSegmentedHitranFolderPath() { return this->baseDownloadFolderPath + "/" + this->segmentedHitranSubFolderName; }
// Folder to contain the segmented HITRAN database files in binary format
QString HitranInputOutput::getSegmentedHitranBinFolderPath() { return this->baseDownloadFolderPath + "/" + this->segmentedHitranSubFolderName + "/" + this->segmentedHitranBinSubFolderName; }

// Method to return the active segmented HITRAN data folder (ASCII or binary)
QString HitranInputOutput::getActiveSegmentedHitranFolderPath()
{
    if (this->hitranFileFormatToUse == 0) { return this->getSegmentedHitranFolderPath(); }
                                     else { return this->getSegmentedHitranBinFolderPath(); }
}

// Method to get the location of the cumulative HITRAN database file
QString HitranInputOutput::getCumulativeHitranFilePath()
{
    // Scanning for the presence of the cumulative HITRAN database file
    scanBaseFolderForCumulativeHitranFile();

    // Returning a complete path to the cumulative HITRAN database file (if present)
    return this->baseDownloadFolderPath + "/" + this->cumulativeHitranFileName;
}

// Method to get and set the variable indicating if the downloaded HITRAN data to be stores as ASCII, binary or both
quint8 HitranInputOutput::getHitranStoreFileFormat() { return this->hitranStoreFileFormat; }
void HitranInputOutput::setHitranStoreFileFormat(quint8 newHitranStoreFileFormat) { this->hitranStoreFileFormat = newHitranStoreFileFormat; }
// Method to get and set the variable indicating if the HITRAN data to be used is ASCII or binary
quint8 HitranInputOutput::getHitranFileFormatToUse() { return this->hitranFileFormatToUse; }
void HitranInputOutput::setHitranFileFormatToUse(quint8 newHitranFileFromatToUse) { this->hitranFileFormatToUse = newHitranFileFromatToUse; }

// Method to scan the base folder for the presence of the cumulative ASCII format HITRAN database file
void HitranInputOutput::scanBaseFolderForCumulativeHitranFile()
{
    QStringList files;  // List of files to be found in a directory
    QFileInfo info;  // Object to get information about a file
    QDir dir;  // directory variable
    int i;     // Loop variable


    // Setting variables related to the download and usage of the HITRAN database ()
    // Getting the list of files already in the folder
    dir.setPath(this->baseDownloadFolderPath);
    files = dir.entryList(QDir::NoDotAndDotDot | QDir::Files);
    if (files.count() == 0) { this->cumulativeHitranFileName = "HITRAN2012.par"; }
    for (i = 0; i < files.count(); i++)
    {
        info.setFile(files[i]);
        if (info.completeSuffix() == "par") {
            this->cumulativeHitranFileName = info.completeBaseName() + "." + info.completeSuffix(); }
        else {
            this->cumulativeHitranFileName = "HITRAN2012.par"; }
    }
}

// Method to get the number of the start and end record corresponding to a given start and end wavenumbers in a single standard 162 ASCII HITRAN database file
void HitranInputOutput::getOneAsciiFileStartEndRecNum(QString fileName, double fromWn, double toWn, quint64 *fromByte, quint64 *toByte)
{
    quint64 startIndex;   // Start byte number in the file
    quint64 endIndex;     // End byte number in the file
    quint64 middleIndex = 0;  // Index of the middle record in the list of records
    quint64 fileBytes;    // Total tumber of bytes in the file
    double prevWn;        // Wavenumber found on the previous step
    QLocale c;       // Object for the data conversion from string to numeric vlaues (locality based)
    QString auxStr;  // Auxiliary string used to extract values from the total line string

    QFile inFile(fileName);  // Name of the HITRAN file in a text form to use
    QString lines;   // Line to read the data from
    HitranOneRec values;     // Values ofor a single record read in from the HITRAN file


    // Opening the HITRAN database text file for reading
    inFile.open(QIODevice::ReadOnly);
    QTextStream in(&inFile);

    // Processing events
    QApplication::processEvents();

    // Getting the total file size in bytes
    fileBytes = inFile.size();

    // Determining the middle index
    startIndex = 0;
    endIndex = floor((double) (fileBytes / 162)) * 162;  // just in case there are extra characters at the end of the file

    // Determining the starting index of the HITRAN dataset to be read in from a file
    prevWn = 1;
    values.setWn(0);
    while (!((prevWn == values.getWn()) || ((endIndex - startIndex) <= 2)))
    {
          // Setting the previously found wavenumber
          prevWn = values.getWn();

          middleIndex = round((double)((startIndex + endIndex) / 324)) * 162;

          // Readin in one line of data
          in.seek(middleIndex);
          lines = in.read(162);

          // Extracting data values
          auxStr = lines.mid(0, 2);
          values.setMolNum(c.toInt(auxStr));  // Molecule number

          // Extracting data values
          convertFromHitranLineString(lines, &values);

          // Processing events
          QApplication::processEvents();

          // Setting new index range
          if (fromWn > values.getWn()) { startIndex = middleIndex; }
          else { endIndex = middleIndex; }
    }
    *fromByte = endIndex;


    // Determining the middle index
    startIndex = 0;
    endIndex = floor((double) (fileBytes / 162)) * 162;  // just in case there are extra characters at the end of the file

    // Determining the starting index of the HITRAN dataset to be read in from a file
    prevWn = 1;
    values.setWn(0);
    while (!((prevWn == values.getWn()) || ((endIndex - startIndex) <= 2)))
    {
        // Setting the previously found wavenumber
        prevWn = values.getWn();

        middleIndex = round((double)((startIndex + endIndex) / 324)) * 162;

        // Readin in one line of data
        in.seek(middleIndex);
        lines = in.read(162);

        // Extracting data values
        convertFromHitranLineString(lines, &values);

        // Processing events
        QApplication::processEvents();

        // Setting new index range
        if (toWn > values.getWn()) { startIndex = middleIndex; }
        else { endIndex = middleIndex; }
    }
    *toByte = endIndex;
}

// Method to get the number of the start and end record corresponding to a given start and end wavenumbers in a single standard 162 ASCII HITRAN database file
void HitranInputOutput::getOneAsciiFileStartEndRecNum(QString fileName, double wn, quint64 *byte)
{
    quint64 startIndex;   // Start byte number in the file
    quint64 endIndex;     // End byte number in the file
    quint64 middleIndex = 0;  // Index of the middle record in the list of records
    quint64 fileBytes;    // Total tumber of bytes in the file
    double prevWn;        // Wavenumber found on the previous step
    QLocale c;       // Object for the data conversion from string to numeric vlaues (locality based)
    QString auxStr;  // Auxiliary string used to extract values from the total line string

    QFile inFile(fileName);  // Name of the HITRAN file in a text form to use
    QString lines;   // Line to read the data from
    HitranOneRec values;     // Values ofor a single record read in from the HITRAN file


    // Opening the HITRAN database text file for reading
    inFile.open(QIODevice::ReadOnly);
    QTextStream in(&inFile);

    // Processing events
    QApplication::processEvents();

    // Getting the total file size in bytes
    fileBytes = inFile.size();

    // Determining the middle index
    startIndex = 0;
    endIndex = floor((double) (fileBytes / 162)) * 162;  // just in case there are extra characters at the end of the file

    // Determining the starting index of the HITRAN dataset to be read in from a file
    prevWn = 1;
    values.setWn(0);
    while (!((prevWn == values.getWn()) || ((endIndex - startIndex) <= 2)))
    {
          // Setting the previously found wavenumber
          prevWn = values.getWn();

          middleIndex = round((double)((startIndex + endIndex) / 324)) * 162;

          // Readin in one line of data
          in.seek(middleIndex);
          lines = in.read(162);

          // Extracting data values
          auxStr = lines.mid(0, 2);
          values.setMolNum(c.toInt(auxStr));  // Molecule number

          // Extracting data values
          convertFromHitranLineString(lines, &values);
          // Only need wavenumber here - replace the above method with a n inline where only the wavenumber gets converted

          // Processing events
          QApplication::processEvents();

          // Setting new index range
          if (wn > values.getWn()) { startIndex = middleIndex; }
          else { endIndex = middleIndex; }
    }
    *byte = endIndex;
}

// Method to get the number of the start and end record corresponding to a given start and end wavenumbers in a single binary format reduced HITRAN database file
void HitranInputOutput::getOneBinFileStartEndRecNum(QString fileName, double fromWn, double toWn, quint64 *fromByte, quint64 *toByte) { ; }

// Method to get the number of the start and end record corresponding to a given start and end wavenumbers in a single binary format reduced HITRAN database file
void HitranInputOutput::getOneBinFileStartEndRecNum(QString fileName, double wn, quint64 *byte)
{
    quint64 middleIndex = 0; // Index of the middle record in the list of records
    quint64 fileBytes;       // Total tumber of bytes in the file
    double prevWn;           // Wavenumber found on the previous step
    QFile inFile(fileName);  // Name of the HITRAN file in a text form to use
    HitranOneRec values;     // Values ofor a single record read in from the HITRAN file
    quint64 startIndex;  // Start byte number in the file
    quint64 endIndex;    // End byte number in the file
    QLocale c;       // Object for the data conversion from string to numeric vlaues (locality based)
    QString auxStr;  // Auxiliary string used to extract values from the total line string
    double currWn;   // Current wavenumber value, cm-1
    quint8 uint8;    // Auxiliary one byte integer
    float auxFloat;
    double auxDouble;


    // Opening the HITRAN database text file for reading
    inFile.open(QIODevice::ReadOnly);
    QDataStream in(&inFile);
    in.setFloatingPointPrecision(QDataStream::DoublePrecision);
    in.device()->reset();

    // Processing events
    QApplication::processEvents();

    // Getting the total file size in bytes
    fileBytes = inFile.size();

    // Determining the middle index
    startIndex = 0;
    endIndex = fileBytes;

    // Determining the starting index of the HITRAN dataset to be read in from a file
    prevWn = 1;
    currWn = 0;
    while (!((prevWn == currWn) || ((endIndex - startIndex) <= 2)))
    {
          // Setting the previously found wavenumber
          prevWn = currWn;

          // Calculating the middle index
          middleIndex = round((double)((startIndex + endIndex) / (2 * Const::hitranOneRecLengthBin))) * Const::hitranOneRecLengthBin;

          // Reading in one line of data
          in.device()->seek(middleIndex);

          // Advancing to the wavenumber and readin it in
          in >> uint8;   // Molecule number
          in >> uint8;   // Isotope number
          in >> currWn;  // Vacuum wavenumber (cm-1)

          // Processing events
          QApplication::processEvents();

          // Setting new index range
          if (wn > currWn) { startIndex = middleIndex; }
          else { endIndex = middleIndex; }
    }
    *byte = endIndex;
}

// Method to set the byte array and convert it to HITRAN variables
void HitranInputOutput::convertIntoHitranRecords(QByteArray *hitranData)
{
    QStringList splitLines;  // HITRAN lines read in from a file
    QString lines;   // String with the data
    int i;   // Loop variable

    QString lineStr;
    QStringRef auxStr;   // Auxiliary string used to extract values from the total line string
    QLocale c(QLocale::English, QLocale::UnitedStates);    // Object for the data conversion from string to numeric vlaues (locality based)
    // The locale has to be set to United states for this one because the HITRAN data was generated in the USA


    // Splitting the read in data into separate lines
    lines = *hitranData;
    splitLines = lines.split("\n", QString::SkipEmptyParts);

    // Converting the read in data from the string representation into the data values
    data.resize(splitLines.count());
    for (i = 0; i < splitLines.count(); i++)
    {
        // Replace this with QStringRef as indicated here, improves the speed a bit
        // https://stackoverflow.com/questions/7584699/qt-get-part-of-qstring

        auxStr = (splitLines[i]).midRef(0, 2);
        (data[i]).setMolNum(c.toInt(auxStr));  // Molecule number
        auxStr = (splitLines[i]).midRef(2, 1);
        // Isotopologue number (handling the HITRAN CO2 isotopologues number 11 and 12 which are denoted as A and B in the *.par format HITRAN database file)
             if (auxStr == "0") { (data[i]).setIsoNum(10); }
        else if (auxStr == "A") { (data[i]).setIsoNum(11); }
        else if (auxStr == "B") { (data[i]).setIsoNum(12); }
                           else { (data[i]).setIsoNum(c.toInt(auxStr)); }
        auxStr = (splitLines[i]).midRef(3, 12);
        (data[i]).setWn(c.toDouble(auxStr));      // Vacuum wavenumber (cm-1)
        auxStr = (splitLines[i]).midRef(15, 10);
        (data[i]).setIntens(c.toDouble(auxStr));  // Line intensity (cm-1 / (molecule * cm^-2) at standard 296K)
        // Need to see if the weighted square of the transition moment and the Einstein A-coefficient (S^-1 are even being used)
        // (data[i]).setR(c.toDouble(auxStr));       // Weighted square of the transition moment (Debye^2) for an electric dipole transition
        // auxStr = (splitLines[i]).midRef(25, 10);
        // (data[i]).setA(c.toDouble(auxStr));       // Einstein A-coefficient (S^-1)
        auxStr = (splitLines[i]).midRef(35, 5);
        (data[i]).setAirHw(c.toDouble(auxStr));   // Air-broadening half-width
        auxStr = (splitLines[i]).midRef(40, 5);
        (data[i]).setSelfHw(c.toDouble(auxStr));  // Self-broadening half-width
        auxStr = (splitLines[i]).midRef(45, 10);
        (data[i]).setLowE(c.toDouble(auxStr));    // Lower state energy (cm-1)
        auxStr = (splitLines[i]).midRef(55, 4);
        (data[i]).setNAir(c.toDouble(auxStr));    // Temperature dependence exponent
        auxStr = (splitLines[i]).midRef(59, 8);
        (data[i]).setAirPShift(c.toDouble(auxStr));  // Air pressure shift (cm^-1 * atm^-1) at 296K

        // extractLineByLineData2004Hitran(splitLines[i], &(data[i]));
    }
}

// Method to read HITRAN records from the cumulative HITRAN file based on the start and end indexes as opposed to wavenumber range
bool HitranInputOutput::readHitranDataAscii(QString filePath, quint64 fromByte, quint64 toByte)
{
    QFile inFile(filePath);  // Name of the HITRAN file in a text form to use
    int i;          // Loop variable
    QString lines;  // Line to read the data from
    QStringList splitLines;  // HITRAN lines read in from a file
    quint64 bytesToRead;  // Number of bytes to read
    bool eofReached;  // Variable indicates if the end of file has been reached

    QString lineStr;
    QStringRef auxStr;   // Auxiliary string used to extract values from the total line string
    QLocale c(QLocale::English, QLocale::UnitedStates);    // Object for the data conversion from string to numeric vlaues (locality based)
    // The locale has to be set to United states for this one because the HITRAN data was generated in the USA


    // Assuming the end of file has not been reached
    eofReached = false;

    // Opening the HITRAN database text file for reading
    if (inFile.open(QIODevice::ReadOnly) == true)
    {
        // Initializing the text stream file
        QTextStream in(&inFile);

        // Reading the middle index HITRAN record
        in.seek(fromByte);

        // Reading in the relevant segment of data from the file
        bytesToRead = inFile.size() - fromByte;
        if (bytesToRead < (toByte - fromByte))
        {
            eofReached = true;
            lines = in.readAll();  // Reading everything available up to the end of file
        }
        else
        {
            lines = in.read(toByte - fromByte);  // Reading only the number of bytes specified
        }

        // Splitting the read in data into separate lines
        splitLines = lines.split("\n", QString::SkipEmptyParts);

        // Converting the read in data from the string representation into the data values
        data.resize(splitLines.count());
        for (i = 0; i < splitLines.count(); i++)
        {
            // Replace this with QStringRef as indicated here, improves the speed a bit
            // https://stackoverflow.com/questions/7584699/qt-get-part-of-qstring

            auxStr = (splitLines[i]).midRef(0, 2);
            (data[i]).setMolNum(c.toInt(auxStr));  // Molecule number
            auxStr = (splitLines[i]).midRef(2, 1);
            // Isotopologue number (handling the HITRAN CO2 isotopologues number 11 and 12 which are denoted as A and B in the *.par format HITRAN database file)
                 if (auxStr == "0") { (data[i]).setIsoNum(10); }
            else if (auxStr == "A") { (data[i]).setIsoNum(11); }
            else if (auxStr == "B") { (data[i]).setIsoNum(12); }
                               else { (data[i]).setIsoNum(c.toInt(auxStr)); }
            auxStr = (splitLines[i]).midRef(3, 12);
            (data[i]).setWn(c.toDouble(auxStr));      // Vacuum wavenumber (cm-1)
            auxStr = (splitLines[i]).midRef(15, 10);
            (data[i]).setIntens(c.toDouble(auxStr));  // Line intensity (cm-1 / (molecule * cm^-2) at standard 296K)
            // Need to see if the weighted square of the transition moment and the Einstein A-coefficient (S^-1 are even being used)
            // (data[i]).setR(c.toDouble(auxStr));       // Weighted square of the transition moment (Debye^2) for an electric dipole transition
			// auxStr = (splitLines[i]).midRef(25, 10);
            // (data[i]).setA(c.toDouble(auxStr));       // Einstein A-coefficient (S^-1)
            auxStr = (splitLines[i]).midRef(35, 5);
            (data[i]).setAirHw(c.toDouble(auxStr));   // Air-broadening half-width
            auxStr = (splitLines[i]).midRef(40, 5);
            (data[i]).setSelfHw(c.toDouble(auxStr));  // Self-broadening half-width
            auxStr = (splitLines[i]).midRef(45, 10);
            (data[i]).setLowE(c.toDouble(auxStr));    // Lower state energy (cm-1)
            auxStr = (splitLines[i]).midRef(55, 4);
            (data[i]).setNAir(c.toDouble(auxStr));    // Temperature dependence exponent
            auxStr = (splitLines[i]).midRef(59, 8);
            (data[i]).setAirPShift(c.toDouble(auxStr));  // Air pressure shift (cm^-1 * atm^-1) at 296K

            // Used to be implemented by calling the below method but this somewhat slows down the operation
            // extractLineByLineData2004Hitran(splitLines[i], &(data[i]));
        }

        // Closing the input file
        inFile.close();
    }

    // Returning a variable indicating if the end of file has been reached
    return eofReached;
}

// Reading records from a certain wavenumber up to a given wavenumber from the HITRAN database
// This function is not used but need to redo this one with a faster calculation approach (inline the data read)
bool HitranInputOutput::readHitranDataAscii(QString filePath, double fromWn, double toWn)
{
    quint64 startIndex;   // Start byte number in the file
    quint64 endIndex;     // End byte number in the file
    quint64 middleIndex = 0;  // Index of the middle record in the list of records
    quint64 fileBytes;    // Total tumber of bytes in the file
    quint64 fromIndex;    // Byte index in a file corresponding to the from Wavenumber
    quint64 toIndex;      // Byte index in a file corresponding to the to wavenumber
    double prevWn;        // Wavenumber found on the previous step
    QLocale c;       // Object for the data conversion from string to numeric vlaues (locality based)
    QString auxStr;  // Auxiliary string used to extract values from the total line string

    QFile inFile(filePath);  // Name of the HITRAN file in a text form to use
    int i;           // Loop variable
    QString lines;   // Line to read the data from
    QStringList splitLines;  // HITRAN lines read in from a file    
    HitranOneRec values;     // Values for a single record read in from the HITRAN file


    // Opening the HITRAN database text file for reading
    inFile.open(QIODevice::ReadOnly);
    QTextStream in(&inFile);

    // Processing events
    QApplication::processEvents();

    // Getting the total file size in bytes
    fileBytes = inFile.size();

    // Determining the middle index
    startIndex = 0;
    endIndex = floor((double) (fileBytes / 162)) * 162;  // just in case there are extra characters at the end of the file

    // Determining the starting index of the HITRAN dataset to be read in from a file
    prevWn = 1;
    values.setWn(0);
    while (!((prevWn == values.getWn()) || ((endIndex - startIndex) <= 2)))
    {
          // Setting the previously found wavenumber
          prevWn = values.getWn();

          middleIndex = round(static_cast<double>((startIndex + endIndex) / 324)) * 162;

          // Readin in one line of data
          in.seek(middleIndex);
          lines = in.read(162);

          // Extracting data values
          auxStr = lines.mid(0, 2);
          values.setMolNum(c.toInt(auxStr));  // Molecule number

          // Extracting data values
          // This needs to be inlined
          convertFromHitranLineString(lines, &values);

          // Processing events
          QApplication::processEvents();

          // Setting new index range
          if (fromWn > values.getWn()) { startIndex = middleIndex; }
          else { endIndex = middleIndex; }
    }
    fromIndex = endIndex;


    // Determining the middle index
    startIndex = 0;
    endIndex = floor((double) (fileBytes / 162)) * 162;  // just in case there are extra characters at the end of the file

    // Determining the starting index of the HITRAN dataset to be read in from a file
    prevWn = 1;
    values.setWn(0);
    while (!((prevWn == values.getWn()) || ((endIndex - startIndex) <= 2)))
    {
        // Setting the previously found wavenumber
        prevWn = values.getWn();

        middleIndex = round((double)((startIndex + endIndex) / 324)) * 162;

        // Readin in one line of data
        in.seek(middleIndex);
        lines = in.read(162);

        // Extracting data values
        convertFromHitranLineString(lines, &values);

        // Processing events
        QApplication::processEvents();

        // Setting new index range
        if (toWn > values.getWn()) { startIndex = middleIndex; }
        else { endIndex = middleIndex; }
    }
    toIndex = endIndex;


    // Reading the middle index HITRAN record
    in.seek(fromIndex);

    // Reading in the relevant segment of data from the file
    lines = in.read(toIndex - fromIndex);

    // Splitting the read in data into separate lines
    splitLines = lines.split("\n", QString::SkipEmptyParts);

    // Converting the read in data from the string representation into the data values
    data.resize(splitLines.count());
    for (i = 0; i < splitLines.count(); i++)
    {
        convertFromHitranLineString(splitLines[i], &(data[i]));
        // strLines.append(splitLines[i]);
    }

    // This should be redone later by determining if the wavenumber requested goes over the file boundaries
    return false;
}

// Method to read HITRAN records from the cumulative HITRAN file based on the start and end indexes as opposed to wavenumber range
bool HitranInputOutput::readHitranDataBin(QString filePath, quint64 fromByte, quint64 toByte)
{
    QFile inFile(filePath);   // Name of the HITRAN file in a text form to use
    int i;                // Loop variable
    quint64 bytesToRead;  // Number of bytes to read
    bool eofReached;      // Variable indicates if the end of file has been reached
    int recsToRead;       // Number of records to read in
    QStringRef auxStr;    // Auxiliary string used to extract values from the total line string
    quint8 uint8;         // Auxiliary 8-bit unsigned integer
    double auxDouble;     // Auxiliary double variable
    float auxFloat;    // Auxiliary float variable


    // Assuming the end of file has not been reached
    eofReached = false;

    // Opening the HITRAN database text file for reading
    if (inFile.open(QIODevice::ReadOnly) == true)
    {
        // Initializing the text stream file
        QDataStream in(&inFile);
        in.device()->reset();
        in.setFloatingPointPrecision(QDataStream::DoublePrecision);

        // Reading the middle index HITRAN record
        in.skipRawData(fromByte);

        // Reading in the relevant segment of data from the file
        bytesToRead = inFile.size() - fromByte;
        if (bytesToRead < (toByte - fromByte))
        {
            recsToRead = bytesToRead / Const::hitranOneRecLengthBin;
            eofReached = true;
        }
        else
        {
            recsToRead = (toByte - fromByte) / Const::hitranOneRecLengthBin;
        }

        // Converting the read in data from the string representation into the data values
        data.resize(recsToRead);
        for (i = 0; i < recsToRead; i++)
        {
            // Molecule number
            in >> uint8;
            (data[i]).setMolNum(uint8);
            // Isotope number
            in >> uint8;
            (data[i]).setIsoNum(uint8);
            // Vacuum wavenumber (cm-1)
            in >> auxDouble;
            (data[i]).setWn(auxDouble);
            // Line intensity (cm-1 / (molecule * cm^-2) at standard 296K)
            in >> auxDouble;
            (data[i]).setIntens(auxDouble);

            in.setFloatingPointPrecision(QDataStream::SinglePrecision);
            // Air-broadening half-width
            in >> auxFloat;
            (data[i]).setAirHw(auxFloat);
            // Self-broadening half-width
            in >> auxFloat;
            (data[i]).setSelfHw(auxFloat);
            in.setFloatingPointPrecision(QDataStream::DoublePrecision);

            // Lower state energy (cm-1)
            in >> auxDouble;
            (data[i]).setLowE(auxDouble);

            in.setFloatingPointPrecision(QDataStream::SinglePrecision);
            // Temperature dependence exponent
            in >> auxFloat;
            (data[i]).setNAir(auxFloat);
            in.setFloatingPointPrecision(QDataStream::DoublePrecision);

            // Air pressure shift (cm^-1 * atm^-1) at 296K
            in >> auxDouble;
            (data[i]).setAirPShift(auxDouble);
        }

        // Closing the input file
        inFile.close();
    }

    // Returning a variable indicating if the end of file has been reached
    return eofReached;
}
bool HitranInputOutput::readHitranDataBin(QString fileName, double fromWn, double toWn)
{
    return true;
}

// Method to read a segment from the segmented ASCII formated HITRAN folder and return new starting points variables
void HitranInputOutput::readSegmentedHitran(quint64 *fromByte, int *fromFileIndex, quint64 downloadBytes)
{
    QString fileName;  // Name of the file to read data from
    QLocale loc;  // Object to convert to numbers from strings
    QDir dir;     // Directory object with the HITRAN segmented files
    int downloadedSegmentNum;   // Segment number of the downloaded region
    QStringList fileNameAndExt;     // Split list with the file name and the extention split up
    QStringList files;      // List of files contained in the directory
    QStringList fileNameSplitList;  // Split name of a file containing the segment number and the spectral range
    int i;  // Loop variable
    bool eofReached;  // Variable indicates if the end of file has been reached


    // Finding the file with the file index required
    // Getting the list of files already in the folder
    dir.setPath(this->getActiveSegmentedHitranFolderPath());
    files = dir.entryList(QDir::NoDotAndDotDot | QDir::Files);

    // Need to re-arrange the files in the list to be in accending file number order
    bubbleSort(&files);

    // Getting the indexes of the files already in place (to determine what needs to be downloaded)
    i = 0;
    fileNameAndExt = files[0].split(".", QString::SkipEmptyParts);
    fileName = fileNameAndExt[0];
    fileNameSplitList = fileName.split("-", QString::SkipEmptyParts);
    downloadedSegmentNum = loc.toInt(fileNameSplitList[0]);
    while (downloadedSegmentNum != *fromFileIndex)
    {
        i++;
        fileNameAndExt = files[i].split(".", QString::SkipEmptyParts);
        fileName = fileNameAndExt[0];
        fileNameSplitList = fileName.split("-", QString::SkipEmptyParts);
        downloadedSegmentNum = loc.toInt(fileNameSplitList[0]);
    }

    // Reading in as many bytes as are available in the first file up to the records step specified or the end of file
    // If the end of file is reached before the number of records in the step is traversed, setting the next pointer to the
    // beginning of the next file for the next read.  This is a simplification procedure.
    if (this->hitranFileFormatToUse == 0) { eofReached = readHitranDataAscii(this->getSegmentedHitranFolderPath() + "/" + fileName + ".par", *fromByte, *fromByte + downloadBytes); }
                                     else { eofReached = readHitranDataBin(this->getSegmentedHitranBinFolderPath() + "/" + fileName + ".bpar", *fromByte, *fromByte + downloadBytes); }

    // Add an if statement here to handle the binary format

    if (eofReached == true)
    {
        *fromFileIndex = *fromFileIndex + 1;
        *fromByte = 0;
    }
    else
    {
        *fromFileIndex = *fromFileIndex;
        *fromByte = *fromByte + downloadBytes;
    }
}

// Bubble sort for the lidar of file name strings
void HitranInputOutput::bubbleSort(QStringList *arr) {

    bool swapped = true;
    int j = 0;
    QString tmp;
    int num;
    int numNext;
    QLocale loc;


    while (swapped)
    {
        swapped = false;
        j++;

        for (int i = 0; i < arr->count() - j; i++)
        {
            num = loc.toInt(((*arr)[i].split("-", QString::SkipEmptyParts))[0]);
            numNext = loc.toInt(((*arr)[i + 1].split("-", QString::SkipEmptyParts))[0]);

            if (num > numNext) {
                tmp = (*arr)[i];
                (*arr)[i] = (*arr)[i + 1];
                (*arr)[i + 1] = tmp;
                swapped = true;
                num = loc.toInt(((*arr)[i].split("-", QString::SkipEmptyParts))[0]);
                numNext = loc.toInt(((*arr)[i + 1].split("-", QString::SkipEmptyParts))[0]);
            }
        }
    }
}

// Method to get the ASCII format HITRAN database start and end indexes and byte numbers in case segmented HITRAN files are in use
void HitranInputOutput::getSegmentedStartEndRec(double fromWn, double toWn, int *fromFileIndex, int *toFileIndex, quint64 *fromByte, quint64 *toByte)
{
    QLocale loc;  // Object to convert to numbers from strings
    QDir dir(this->getSegmentedHitranFolderPath());   // Directory object with the HITRAN segmented files
    int downloadedSegmentNum;   // Segment number of the downloaded region
    double segmentStartWn;  // Rounded to integer start wavenumber
    double segmentEndWn;    // Rounded to integer end wavenumber
    QStringList files;      // List of files contained in the directory
    QStringList fileNameSplitList;  // Split name of a file containing the segment number and the spectral range
    int i;  // Loop variable


    // If the binary HITRAN data files are to be used in calculations, adding the BIN sub-directory
    if (this->hitranFileFormatToUse == 1) { dir.setPath(this->getSegmentedHitranBinFolderPath()); }

    // Getting the list of files already in the folder
    files = dir.entryList(QDir::NoDotAndDotDot | QDir::Files);

    // Rearranging the files to be in accending order in terms of their numbering (each file is numbered within the file name)
    bubbleSort(&files);

    // Removing the file extension
    for (i = 0; i < files.count(); i++)
    {
        files[i] = (files[i].split("."))[0];
    }

    // Getting the indexes of the files already in place (to determine what needs to be downloaded)
    i = 0;
    fileNameSplitList = files[0].split("-", QString::SkipEmptyParts);
    downloadedSegmentNum = loc.toInt(fileNameSplitList[0]);
    segmentStartWn = loc.toInt(fileNameSplitList[1]);
    segmentEndWn = loc.toInt(fileNameSplitList[2]);
    while (!((fromWn >= segmentStartWn) && (fromWn <= segmentEndWn)))
    {        
        fileNameSplitList = files[i].split("-", QString::SkipEmptyParts);
        downloadedSegmentNum = loc.toInt(fileNameSplitList[0]);
        segmentStartWn = loc.toInt(fileNameSplitList[1]);
        segmentEndWn = loc.toInt(fileNameSplitList[2]);
        i++;
    }
    if (i > 0) { i--; }

    // Finding the record numbers within the files (Binary or ASCII files)
    *fromFileIndex = downloadedSegmentNum;
    if (this->hitranFileFormatToUse == 0) { getOneAsciiFileStartEndRecNum(dir.path() + "/" + files[i] + ".par", fromWn, fromByte); }
                                     else { getOneBinFileStartEndRecNum(dir.path() + "/" + files[i] + ".bpar", fromWn, fromByte); }

    i = 1;
    fileNameSplitList = files[0].split("-", QString::SkipEmptyParts);
    downloadedSegmentNum = loc.toInt(fileNameSplitList[0]);
    segmentStartWn = loc.toInt(fileNameSplitList[1]);
    segmentEndWn = loc.toInt(fileNameSplitList[2]);

    while (!((toWn >= segmentStartWn) && (toWn <= segmentEndWn)))
    {
        fileNameSplitList = files[i].split("-", QString::SkipEmptyParts);
        downloadedSegmentNum = loc.toInt(fileNameSplitList[0]);
        segmentStartWn = loc.toInt(fileNameSplitList[1]);
        segmentEndWn = loc.toInt(fileNameSplitList[2]);
        i++;
    }    
    if (i > 0) { i--; }
    *toFileIndex = i;

    // Finding the record numbers within the files (in ASCII or binary format)
    *toFileIndex = downloadedSegmentNum;
    if (this->hitranFileFormatToUse == 0) { getOneAsciiFileStartEndRecNum(dir.path() + "/" + files[i] + ".par", toWn, toByte); }
                                     else { getOneBinFileStartEndRecNum(dir.path() + "/" + files[i] + ".bpar", toWn, toByte); }
}

// Method to get the number of records for segmented HITRAN files based on the start and end index and file number
quint64 HitranInputOutput::getNumSegmentedRecsToProcess(int fromFileIndex, int toFileIndex, quint64 fromByte, quint64 toByte)
{
    QLocale loc;         // Object to convert to numbers from strings
    QDir dir(this->getSegmentedHitranFolderPath());   // Directory object with the HITRAN segmented files
    QFileInfo info;      // Object to get the file information
    int downloadedSegmentNum;   // Segment number of the downloaded region
    QStringList files;   // List of files contained in the directory
    QStringList fileNameSplitList;   // Split name of a file containing the segment number and the spectral range
    QString ext;         // File extention (par for ASCII, bpar for binary)
    int recordLength;    // Length of one record (162 for ASCII, )
    int i;               // Loop variable
    quint64 numRecs;     // Number of records


    // Changing the directory location and the extention if the binary representation of the HITRAN database is used
    recordLength = Const::hitranOneRecLengthAscii;
    ext = "par";
    if (this->hitranFileFormatToUse == 1)
    {
        dir.setPath(this->getSegmentedHitranBinFolderPath());
        recordLength = Const::hitranOneRecLengthBin;
        ext = "bpar";
    }

    // Getting the list of files already in the folder
    files = dir.entryList(QDir::NoDotAndDotDot | QDir::Files);

    // Removing the file extension
    for (i = 0; i < files.count(); i++)
    {
        files[i] = (files[i].split("."))[0];
    }

    // Calculating the total size in all files and the number of records
    numRecs = 0;
    for (i = 0; i < files.count(); i++)
    {
        fileNameSplitList = files[i].split("-", QString::SkipEmptyParts);
        downloadedSegmentNum = loc.toInt(fileNameSplitList[0]);
        if ((downloadedSegmentNum >= fromFileIndex) && (downloadedSegmentNum <= (toFileIndex - 1)))
        {
            info.setFile(dir.path() + "/" + files[i] + "." + ext);
            numRecs = numRecs + info.size();
        }
    }
    numRecs = numRecs + toByte - fromByte;
    numRecs = numRecs / recordLength;

    return numRecs;
}

// Method to save HITRAN data into a binary file
void HitranInputOutput::convertIntoHitranBinFile(QString fileName, QByteArray *hitranData)
{
    QStringList splitLines;  // HITRAN lines read in from a file
    QString lines;   // String with the data
    quint8 auxInt;   // Auxiliary 8-bit unsigned integer
    float auxFloat;  // Auxiliary floating point number
    int i;  // Loop variable

    QStringRef auxStr;   // Auxiliary string used to extract values from the total line string
    QLocale c(QLocale::English, QLocale::UnitedStates);    // Object for the data conversion from string to numeric vlaues (locality based)
    // The locale has to be set to United states for this one because the HITRAN data was generated in the USA


    // Name of the HITRAN file in a text form with segmented information
    QFile outFile(this->getSegmentedHitranBinFolderPath() + "/" + fileName);

    // Splitting the read in data into separate lines
    lines = QString::fromLatin1(*hitranData);
    splitLines = lines.split("\n", QString::SkipEmptyParts);

    // Opening the HITRAN database binary file for writing
    outFile.open(QIODevice::ReadWrite);
    QDataStream out(&outFile);
    // Setting the QDataStream floating point precision
    out.setFloatingPointPrecision(QDataStream::DoublePrecision);

    // Converting the read in data from the string representation into the data values
    for (i = 0; i < splitLines.count(); i++)
    {
        // Replace this with QStringRef as indicated here, improves the speed a bit
        // https://stackoverflow.com/questions/7584699/qt-get-part-of-qstring

        // Molecule number
        auxStr = (splitLines[i]).midRef(0, 2);
        auxInt = c.toUInt(auxStr);
        out << auxInt;
        // Isotope number
        auxStr = (splitLines[i]).midRef(2, 1);
        // Isotopologue number (handling the HITRAN CO2 isotopologues number 11 and 12 which are denoted as A and B in the *.par format HITRAN database file)
             if (auxStr == "0") { auxInt = 10; }
        else if (auxStr == "A") { auxInt = 11; }
        else if (auxStr == "B") { auxInt = 12; }
                           else { auxInt = c.toUInt(auxStr); }
        out << auxInt;
        // Vacuum wavenumber (cm-1)
        auxStr = (splitLines[i]).midRef(3, 12);
        out << c.toDouble(auxStr);
        // Line intensity (cm-1 / (molecule * cm^-2) at standard 296K)
        auxStr = (splitLines[i]).midRef(15, 10);
        out << c.toDouble(auxStr);

        // Need to see if the weighted square of the transition moment and the Einstein A-coefficient (S^-1 are even being used)
        // (data[i]).setR(c.toDouble(auxStr));       // Weighted square of the transition moment (Debye^2) for an electric dipole transition
        // auxStr = (splitLines[i]).midRef(25, 10);
        // (data[i]).setA(c.toDouble(auxStr));       // Einstein A-coefficient (S^-1)

        out.setFloatingPointPrecision(QDataStream::SinglePrecision);
        // Air-broadening half-width
        auxStr = (splitLines[i]).midRef(35, 5);
        auxFloat = c.toFloat(auxStr);
        out << auxFloat;
        // Self-broadening half-width
        auxStr = (splitLines[i]).midRef(40, 5);
        auxFloat = c.toFloat(auxStr);
        out << auxFloat;
        out.setFloatingPointPrecision(QDataStream::DoublePrecision);

        // Lower state energy (cm-1)
        auxStr = (splitLines[i]).midRef(45, 10);
        out << c.toDouble(auxStr);

        out.setFloatingPointPrecision(QDataStream::SinglePrecision);
        // Temperature dependence exponent
        auxStr = (splitLines[i]).midRef(55, 4);
        auxFloat = c.toFloat(auxStr);
        out << auxFloat;
        out.setFloatingPointPrecision(QDataStream::DoublePrecision);

        // Air pressure shift (cm^-1 * atm^-1) at 296K
        auxStr = (splitLines[i]).midRef(59, 8);
        out << c.toDouble(auxStr);
    }

    // Closing the output file
    outFile.close();
}

// Method to get the array with the read in HITRAN data
QVector<HitranOneRec> *HitranInputOutput::getHitranRecords() { return &(this->data); }

// Function to extract line-by-line data for a single spectral line
void HitranInputOutput::convertFromHitranLineString(QString lineStr, HitranOneRec *lineData) {

    QString auxStr;  // Auxiliary string used to extract values from the total line string
    QLocale c(QLocale::English, QLocale::UnitedStates);    // Object for the data conversion from string to numeric vlaues (locality based)
    // The locale has to be set to United states for this one because the HITRAN data was generated in the USA


    auxStr = lineStr.mid(0, 2);
    lineData->setMolNum(c.toInt(auxStr));  // Molecule number
    auxStr = lineStr.mid(2, 1);
    // Isotopologue number (handling the HITRAN CO2 isotopologues number 11 and 12 which are denoted as A and B in the *.par format HITRAN database file)
         if (auxStr == "0") { lineData->setIsoNum(10); }
    else if (auxStr == "A") { lineData->setIsoNum(11); }
    else if (auxStr == "B") { lineData->setIsoNum(12); }
                       else { lineData->setIsoNum(c.toInt(auxStr)); }
    auxStr = lineStr.mid(3, 12);
    lineData->setWn(c.toDouble(auxStr));      // Vacuum wavenumber (cm-1)
    auxStr = lineStr.mid(15, 10);
    lineData->setIntens(c.toDouble(auxStr));  // Line intensity (cm-1 / (molecule * cm^-2) at standard 296K)
    lineData->setR(c.toDouble(auxStr));       // Weighted square of the transition moment (Debye^2) for an electric dipole transition
    auxStr = lineStr.mid(25, 10);
    lineData->setA(c.toDouble(auxStr));       // Einstein A-coefficient (S^-1)
    auxStr = lineStr.mid(35, 5);
    lineData->setAirHw(c.toDouble(auxStr));   // Air-broadening half-width
    auxStr = lineStr.mid(40, 5);
    lineData->setSelfHw(c.toDouble(auxStr));  // Self-broadening half-width
    auxStr = lineStr.mid(45, 10);
    lineData->setLowE(c.toDouble(auxStr));    // Lower state energy (cm-1)
    auxStr = lineStr.mid(55, 4);
    lineData->setNAir(c.toDouble(auxStr));    // Temperature dependence exponent
    auxStr = lineStr.mid(59, 8);
    lineData->setAirPShift(c.toDouble(auxStr));  // Air pressure shift (cm^-1 * atm^-1) at 296K

    // These additional fields are to be implemented later
    //    auxStr = lineStr.mid(67, 15);
    //    lineData->setUpperV(auxStr);    // Upper state global quanta
    //    auxStr = lineStr.mid(82, 15);
    //    lineData->setLowerV(auxStr);    // Lower state global quanta
    //    auxStr = lineStr.mid(197, 15);
    //    lineData->setUpperQ(auxStr);    // Upper state local quanta
    //    auxStr = lineStr.mid(112, 15);
    //    lineData->setLowerQ(auxStr);    // Lower state local quanta
    //    auxStr = lineStr.mid(127, 6);
    //    lineData->setIntUncWn(c.toInt(auxStr));      // Wavenumber uncertainty index
    //    auxStr = lineStr.mid(133, 12);
    //    lineData->setIntUncIntens(c.toInt(auxStr));  // Intensity uncertainty index
    //    auxStr = lineStr.mid(145, 1);
    //    lineData->setIntUncAirHw(c.toInt(auxStr));   // Air-broadening half-width uncertainty index
    //    auxStr = lineStr.mid(146, 7);
    //    lineData->setIntUncSelfHw(c.toInt(auxStr));  // Self-broadening half-width uncertainty index
    //    auxStr = lineStr.mid(153, 7);
    //    lineData->setIntUncNAir(c.toInt(auxStr));    // Temperature dependent exponent uncertainty index

    //    The values below are obtained from the above read information and additional datafiles
    //    lineData->setIntUncAirPShift(0);   // Air-pressure shift uncertainty index
    //    lineData->setDblUncWn(0);      // Wavenumber uncertainty value
    //    lineData->setDblUncIntens(0);  // Intensity uncertainty value
    //    lineData->setDblUncAirHw(0);   // Air-broadening half-width uncertainty value
    //    lineData->setDblUncSelfHw(0);  // Self-broadening half-width uncertainty value
    //    lineData->setDblUncNAir(0);    // Temperature dependent exponent uncertainty value
    //    lineData->setDblUncAirPShift(0);  // Air-pressure shift uncertainty value
    //    lineData->setRefWn(0);      // Wavenumber uncertainty reference article number
    //    lineData->setRefIntens(0);  // Intensity uncertainty reference article number
    //    lineData->setRefAirHw(0);   // Air-broadening half-width uncertainty reference article number
    //    lineData->setRefSelfHw(0);  // Self-broadening half-width uncertainty reference article number
    //    lineData->setRefNAir(0);    // Temperature dependent exponent uncertainty reference article number
    //    lineData->setRefAirPShift(0);  // Air-pressure shift uncertainty reference article number
    //    lineData->setLMix("");   // Availablility of programs and data for the case of line-mixing
    //    lineData->setGUpper(0);  // Statistical weight of the upper state
    //    lineData->setGLower(0);  // Statistical weight of the lower state
}

// Method to get the string lines read in from the HITRAN database (no conversion to numerical representation)
QVector<QString> *HitranInputOutput::getStrLines()
{
    return &(this->strLines);
}

// Method to save data to file (usually appending at the right place)
void HitranInputOutput::saveStringLinesToFile(QString filePath, QString *data)
{
    QFile hitranFile(filePath);

    // Need to find the location within the file where the data is to be saved
    double fromWn;        // Wavenumber to find
    quint64 startIndex;   // Start byte number in the file
    quint64 endIndex;     // End byte number in the file
    quint64 middleIndex = 0;  // Index of the middle record in the list of records
    quint64 fileBytes;    // Total tumber of bytes in the file
    quint64 fromIndex;    // Byte index in a file corresponding to the from Wavenumber
    double prevWn;        // Wavenumber found on the previous step
    QLocale c;       // Object for the data conversion from string to numeric vlaues (locality based)
    QString auxStr;  // Auxiliary string used to extract values from the total line string
    QString line;   // Line to read the data from
    HitranOneRec values;  // Values ofor a single record read in from the HITRAN file
    QString oneRecStr;    // String to contain a single HITRAN database record
    QMessageBox msgBox;   // Message box to display errors to the user
    QString buffer;       // Buffer to read in selected portion of a file to insert text data into a file


    // Need to work with data read in from a file for the entire spectral range in question and then insert
    // this segment back into the file after the insersion of the downloaded regions has been carried out

    // Maybe better algorithm could be written so that it is not necessary to separately hande the very first and the very last indices
    if (hitranFile.open(QIODevice::ReadWrite))
    {
        QTextStream ioStream(&hitranFile);

        // Getting the total file size in bytes
        fileBytes = hitranFile.size();

        // Determining the fromWn wavenumber from the data string to save to a file
        oneRecStr = data->mid(0, 162);
        convertFromHitranLineString(oneRecStr, &values);
        fromWn = values.getWn();

        // Determining the middle index
        startIndex = 0;
        endIndex = ceil((double) (fileBytes / 162)) * 162;  // Just in case there are extra characters at the end of the file

        // Determining the starting index of the HITRAN dataset to be read in from a file
        prevWn = 1;
        values.setWn(0);
        while (!((prevWn == values.getWn()) || ((endIndex - startIndex) <= 2)))
        {
            // Setting the previously found wavenumber
            prevWn = values.getWn();

            // Determining the mid index in the dataset
            middleIndex = floor((double)((startIndex + endIndex) / 324)) * 162;

            // Reading in one line of data
            ioStream.seek(middleIndex);
            line = ioStream.read(162);
            convertFromHitranLineString(line, &values);

            // Setting new index range
            if (fromWn > values.getWn()) { startIndex = middleIndex; }
            else { endIndex = middleIndex; }
        }
        // Probably need to move to the fromIndex location (carry out the seek operation one more time)
        fromIndex = endIndex;

        // Seeking the location in the file to write the data out at
        ioStream.seek(fromIndex);

        // Reading in the trailing data from the file (used to insert the text in the middle of a file)
        buffer = ioStream.read(hitranFile.size() - fromIndex);

        // Saving the data to a file at a specified position
        ioStream.seek(fromIndex); // Returning back from the previous position
        ioStream << (*data);
        ioStream << buffer;

        // Closing the file after the writing has been completed
        hitranFile.close();
    }
    else
    {
        msgBox.setText("Error opening HITRAN database for writing.");
        msgBox.exec();
    }
}
