// bytran -|- spectral calculations for portable devices using the HITRAN database
//
// Developed by Dzianis Pliutau (Denis Plutov) and Kostya Roslyakov
// www.bytran.org | bytran@bytran.org
// Copyright (c) 2016 Dzianis Pliutau
// Distributed under the MIT license
//
// The MIT License (MIT)
// Copyright (c) 2016 Dzianis Pliutau
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation
// files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,
// modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software
// is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
// OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
// LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
// IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.



// Object to read/write HITRAN lines within a specified spectral range
#ifndef HITRANINPUTOUTPUT_H
#define HITRANINPUTOUTPUT_H

#include <QVector>
#include "hitranonerec.h"
#include "spectrum.h"
#include "hitranonemol.h"
#include <QByteArray>

class HitranInputOutput
{

public:

    HitranInputOutput();

    // HITRAN database local storage location related methods
    // Name of the cumulative HITRAN database file (the one which contains all records, old *.par cumulative file style)
    QString getCumulativeHitranFileName();
    void setCumulativeHitranFileName(QString newCumulativeHitranFileName);
    // Method to get the location of the cumulative HITRAN database file
    QString getCumulativeHitranFilePath();

    // Folder with the cumulative HITRAN database file (where the complete HITRAN database file is to be stored), this folder will also contain a sub-folder with the HITRAN downloaded in segments from HITRANOnline
    QString getBaseDownloadFolderPath();
    void setBaseDownloadFolderPath(QString newBaseDowloadFolderPath);

    // Method will check if the ASCII and binary HITRAN folders exist and create them if the do not exist
    void createHitranDataDirsIfNoneExist();

    // Name of the folder to contain the segmented HITRAN database files in ASCII format
    QString getSegmentedHitranSubFolderName();
    void setSegmentedHitranSubFolderName(QString newSegmentedHitranSubFolderName);
    // Name of the folder to contain the segmented HITRAN database files in binary format
    QString getSegmentedHitranBinSubFolderName();
    void setSegmentedHitranBinSubFolderName(QString newSegmentedHitranBinSubFolderName);

    // Folder to contain the segmented HITRAN database files in ASCII format
    QString getSegmentedHitranFolderPath();
    // Folder to contain the segmented HITRAN database files in binary format
    QString getSegmentedHitranBinFolderPath();
    // Returns the segmented HITRAN folder currently set as active
    QString getActiveSegmentedHitranFolderPath();

    // Method to get and set the variable indicating if the downloaded HITRAN data to be stores as ASCII, binary or both
    quint8 getHitranStoreFileFormat();
    void setHitranStoreFileFormat(quint8 newHitranDataStoreFormat);
    // Method to get and set the variable indicating if the HITRAN data to be used is ASCII or binary
    quint8 getHitranFileFormatToUse();
    void setHitranFileFormatToUse(quint8 newHitranDataFromatToUse);

    // Method to scan the base folder for the presence of the cumulative ASCII format HITRAN database file
    void scanBaseFolderForCumulativeHitranFile();

    // Methods to get the number of the start and end record corresponding to a given start and end wavenumbers in a single standard 162 ASCII HITRAN database file
    void getOneAsciiFileStartEndRecNum(QString fileName, double fromWn, double toWn, quint64 *fromByte, quint64 *toByte);
    void getOneAsciiFileStartEndRecNum(QString fileName, double wn, quint64 *byte);
    // Methods to get the number of the start and end record corresponding to a given start and end wavenumbers in a single binary format reduced HITRAN database file
    void getOneBinFileStartEndRecNum(QString fileName, double fromWn, double toWn, quint64 *fromByte, quint64 *toByte);
    void getOneBinFileStartEndRecNum(QString fileName, double wn, quint64 *byte);

    // Methods to read HITRAN records from an ASCII HITRAN database file based on the start and end indexes or wavenumbers
    bool readHitranDataAscii(QString filePath, quint64 fromByte, quint64 toByte);
    bool readHitranDataAscii(QString fileName, double fromWn, double toWn);
    // Methods to read HITRAN records from a binary HITRAN database file based on the start and end indexes or wavenumbers
    bool readHitranDataBin(QString filePath, quint64 fromByte, quint64 toByte);
    bool readHitranDataBin(QString fileName, double fromWn, double toWn);

    // Method to get the segmented HITRAN folder start and end byte indexes and file numbers (handles both ASCII as well as the binary format)
    void getSegmentedStartEndRec(double fromWn, double toWn, int *fromFileIndex, int *toFileIndex, quint64 *fromByte, quint64 *toByte);

    // Method to read a segment from the segmented HITRAN folder and return new starting points variables
    void readSegmentedHitran(quint64 *fromByte, int *fromFileIndex, quint64 downloadBytes);

    // Method to get the number of records for segmented HITRAN files based on the start and end index and file number
    quint64 getNumSegmentedRecsToProcess(int fromFileIndex, int toFileIndex, quint64 fromByte, quint64 toByte);

    // Method to save HITRAN data into a binary file
    void convertIntoHitranBinFile(QString fileName, QByteArray *hitranData);

    // Method to get the string lines read in from the HITRAN database (no conversion to numerical representation)
    QVector<QString> *getStrLines();

    // Method to save data to file (usually appending at the right place)
    void saveStringLinesToFile(QString filePath, QString *data);

    // Method to set the byte array and convert it to HITRAN variables
    void convertIntoHitranRecords(QByteArray *hitranData);

    // Method to get the array with the read in HITRAN data
    QVector<HitranOneRec> *getHitranRecords();

private:

    // Function to extract line-by-line data for a single spectral line
    void convertFromHitranLineString(QString lineStr, HitranOneRec *lineData);

    // Bubble sort for the lidar of file name strings
    void bubbleSort(QStringList *arr);

    // Variable with the HITRAN data
    QVector<HitranOneRec> data;

    // HITRAN lines
    QVector<QString> strLines;

    // Variables related to the HITRAN database storage location (total HITRAN or segmented)
    QString cumulativeHitranFileName;   // Name of the file to save the downloaded data to (cumulative HITRAN file name)
    QString baseDownloadFolderPath;     // Base download path where the HITRAN database is to be stored
    QString segmentedHitranSubFolderName;     // Folder to contain the segmented HITRAN database files in the ASCII format and the subfolder for the binary format files
    QString segmentedHitranBinSubFolderName;  // Subfolder to contain segmented HITRAN files in a binary format

    // Variable indicating if the downloaded HITRAN data to be stores as ASCII, binary or both
    quint8 hitranStoreFileFormat;    // Store downloaded HITRAN files as: 0 - ASCII, 1 - binary, 2 - ASCII and binary
    // Variable indicating if the HITRAN data to be used is ASCII or binary
    quint8 hitranFileFormatToUse;    // Use HITRAN saved in one of the following formats: 0 - ASCII, 1 - binary    

};

#endif // HITRANINPUTOUTPUT_H
