// bytran -|- spectral calculations for portable devices using the HITRAN database
//
// Developed by Dzianis Pliutau (Denis Plutov) and Kostya Roslyakov
// www.bytran.org | bytran@bytran.org
// Copyright (c) 2016 Dzianis Pliutau
// Distributed under the MIT license
//
// The MIT License (MIT)
// Copyright (c) 2016 Dzianis Pliutau
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation
// files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,
// modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software
// is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
// OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
// LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
// IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.



// Object is used to relay the pathlength information betweeen the C++ data objects and the QML interface
#include "pathlengthqml.h"
#include "auxiliary.h"
#include <QMessageBox>
#include <QLocale>
#include <QString>
#include <QLocale>

PathlengthQml::PathlengthQml(QObject *parent, BytranCore *inCore) : QObject(parent)
{
    double pathMeters;  // Pathlength in meter units
    double pathCurrUnits;  // Pathlength in current units


    // So that commas are not inserted into numbers
    loc.setNumberOptions(QLocale::OmitGroupSeparator);

    // Passing in the core object
    core = inCore;

    // Passing variables from the C++ classes into the QML transfer object
    pathMeters = core->getParams()->getHorizParams()->getPathLength();
    pathCurrUnits = core->getPathlengthUnitsConverter()->fromDefaultToCurrent(pathMeters);
    m_pathlength = loc.toString(pathCurrUnits, 'g', 6);
    m_units = core->getPathlengthUnitsConverter()->getUnits();

    // Connections
    connect(this->core, SIGNAL(valuesUpdated()), this, SLOT(setUpdatedValues()));
}

// Slot is used to update all values in the corresponding QML dialog
void PathlengthQml::setUpdatedValues()
{
    emit pathlengthChanged();
    emit unitsChanged();
}

PathlengthQml::PathlengthQml(const QString &pathlength, const int &units, QObject *parent)
    : QObject(parent), m_pathlength(pathlength), m_units(units)
{
}

QString PathlengthQml::pathlength() const
{
    double pathMeters;  // Pathlength in meter units
    double pathCurrUnits;  // Pathlength in current units

    pathMeters = core->getParams()->getHorizParams()->getPathLength();
    pathCurrUnits = core->getPathlengthUnitsConverter()->fromDefaultToCurrent(pathMeters);
    return loc.toString(pathCurrUnits, 'g', 6);
}
void PathlengthQml::setPathlength(const QString &pathlength)
{
    double pathDefaultUnits;  // Temperature in current units to be converted to Kelvin

    if ((pathlength != m_pathlength) && Auxiliary::acceptInput(pathlength)) {
        m_pathlength = pathlength;
        // Only the horizontal path mode is available at this point
        pathDefaultUnits = core->getPathlengthUnitsConverter()->fromCurrentToDefault(loc.toDouble(m_pathlength));
        core->getParams()->getHorizParams()->setPathLength(pathDefaultUnits);
        emit pathlengthChanged();
    }
}

quint8 PathlengthQml::units() const
{
    return core->getPathlengthUnitsConverter()->getUnits();
}
void PathlengthQml::setUnits(const quint8 &units)
{
    double pathMeters;     // Temperature in Kelvin units
    double pathCurrUnits;  // Temperature in current units

    if (units != m_units) {
        m_units = units;
        // Saving the units into the object used to convert to different pathlength measurement units
        core->getPathlengthUnitsConverter()->setUnits(m_units);

        // Updating the displayed pathlength
        pathMeters = core->getParams()->getHorizParams()->getPathLength();
        pathCurrUnits = core->getPathlengthUnitsConverter()->fromDefaultToCurrent(pathMeters);
        m_pathlength = loc.toString(pathCurrUnits, 'g', 6);

        emit unitsChanged();
        emit pathlengthChanged();
    }
}
