// bytran -|- spectral calculations for portable devices using the HITRAN database
//
// Developed by Dzianis Pliutau (Denis Plutov) and Kostya Roslyakov
// www.bytran.org | bytran@bytran.org
// Copyright (c) 2016 Dzianis Pliutau
// Distributed under the MIT license
//
// The MIT License (MIT)
// Copyright (c) 2016 Dzianis Pliutau
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation
// files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,
// modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software
// is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
// OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
// LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
// IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.



// Object is used to relay the total pressure information betweeen the C++ data objects and the QML interface
#include "totalpressqml.h"
#include "auxiliary.h"
#include <QLocale>
#include <QMessageBox>

TotalPressQml::TotalPressQml(QObject *parent, BytranCore *inCore) : QObject(parent)
{
    double totalPressAtm;        // Total pressure in default internal units (atm)
    double totalPressCurrUnits;  // Total pressure in current pressure units


    // So that commas are not inserted into numbers
    loc.setNumberOptions(QLocale::OmitGroupSeparator);

    // Passing in the core object
    core = inCore;

    // Saving the total pressure and units value into a local variable
    totalPressAtm = core->getParams()->getHorizParams()->getTotalPress();
    totalPressCurrUnits = core->getConcPressUnitsConverter()->fromDefaultToCurrentPressUnits(totalPressAtm);
    m_totalPressure = loc.toString(totalPressCurrUnits, 'g', 6);
    m_units = core->getConcPressUnitsConverter()->getPressUnits();

    // Connections
    connect(core->getSensors(), SIGNAL(totalPressureUpdated()), this, SLOT(setUpdatedValues()));
    connect(this->core, SIGNAL(valuesUpdated()), this, SLOT(setUpdatedValues()));
}

// Slot is used to update all values in the corresponding QML dialog
void TotalPressQml::setUpdatedValues()
{
    emit totalPressureChanged();
    emit unitsChanged();
}
void TotalPressQml::updateTotalPress() { setUpdatedValues(); }

TotalPressQml::TotalPressQml(const QString &totalPressure, const int &units, QObject *parent)
    : QObject(parent), m_totalPressure(totalPressure), m_units(units)
{
}

QString TotalPressQml::totalPressure() const
{
    double totalPressAtm;        // Total pressure in default internal units (atm)
    double totalPressCurrUnits;  // Total pressure in current pressure units

    totalPressAtm = core->getParams()->getHorizParams()->getTotalPress();
    totalPressCurrUnits = core->getConcPressUnitsConverter()->fromDefaultToCurrentPressUnits(totalPressAtm);
    return loc.toString(totalPressCurrUnits, 'g', 6);
}
void TotalPressQml::setTotalPressure(const QString &totalPressure)
{
    double totalPressDefaultUnits;  // Total pressure in current units to be converted to atm

    if ((totalPressure != m_totalPressure) && Auxiliary::acceptInput(totalPressure)) {
        m_totalPressure = totalPressure;
        // Need to convert the value to default units (mbar)
        totalPressDefaultUnits = core->getConcPressUnitsConverter()->fromCurrentToDefaultPressUnits(loc.toDouble(m_totalPressure));
        core->getParams()->getHorizParams()->setTotalPress(totalPressDefaultUnits);
        emit totalPressureChanged();
    }
}

quint8 TotalPressQml::units() const
{
    return core->getConcPressUnitsConverter()->getPressUnits();
}
void TotalPressQml::setUnits(const quint8 &units)
{
    double totalPressAtm;  // Total pressure in atm units
    double totalPressCurrUnits;  // Total pressure in current units

    if (units != m_units) {
        m_units = units;

        // Saving the units into the object used to convert to different pathlength measurement units
        core->getConcPressUnitsConverter()->setPressUnits(m_units);

        // Updating the displayed temperature
        totalPressAtm = core->getParams()->getHorizParams()->getTotalPress();
        totalPressCurrUnits = core->getConcPressUnitsConverter()->fromDefaultToCurrentPressUnits(totalPressAtm);
        m_totalPressure = loc.toString(totalPressCurrUnits, 'g', 6);

        // Emitting signals notifying about the change in values and units
        emit unitsChanged();
        emit totalPressureChanged();
    }    
}
