// bytran -|- spectral calculations for portable devices using the HITRAN database
//
// Developed by Dzianis Pliutau (Denis Plutov) and Kostya Roslyakov
// www.bytran.org | bytran@bytran.org
// Copyright (c) 2016 Dzianis Pliutau
// Distributed under the MIT license
//
// The MIT License (MIT)
// Copyright (c) 2016 Dzianis Pliutau
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation
// files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,
// modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software
// is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
// OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
// LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
// IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.



// Object stores curently active wavelength units and performs conversion from / to the default wavelength units to the current wavelength units
#include "wavelength.h"
#include "math.h"
#include <QMessageBox>
#include <QLocale>

Wavelength::Wavelength()
{
    this->axisTitles.resize(4);
    this->axisTitles[0] = "Wavelength, nm";
    this->axisTitles[1] = "Wavelength, micron";
    this->axisTitles[2] = "Wavenumber, cm-1";
    this->axisTitles[3] = "Frequency, THz";

    // Default parameters if not initialized;
    this->currentUnits = 1;
    this->startWn = 2000;
    this->endWn = 2300;
    this->wnResolMode = 1;
    this->wnResol = 0.01;
    this->numPoints = 10000;
    this->enableRecalcOfCalculatedValues = true;
}

// Method to get and set the current total pressure units
quint8 Wavelength::getUnits() { return this->currentUnits; }
void Wavelength::setUnits(quint8 newUnits) { this->currentUnits = newUnits; }

// Method to convert from default units to current units
double Wavelength::fromDefaultToCurrent(double defaultValue)
{
    // 0 - Wavelength, nm;  1 - Wavelength, micron, 2 - Wavenumber, cm-1;  3 - Frequency, THz;  ( 4 - Attenuation, Db try to add later, this is not dB but something involving Db, need to check )
    switch(currentUnits)
    {
        case 0:
            return 10000000 / defaultValue;
            break;
        case 1:
            return 10000 / defaultValue;
            break;
        case 2:
            return defaultValue;
            break;
        case 3:
            return defaultValue * 29979245800 * 0.000000000001;   // 100 is added in to take into account the conversion from centimeters to meters, the other multiplier 1.0e12 is to use THz
            break;
        default:
            return defaultValue;
            break;
    }
}

// Method to convert from the current units to the default units
double Wavelength::fromCurrentToDefault(double currentValue)
{
    // 0 - Wavelength, nm;  1 - Wavelength, micron, 2 - Wavenumber, cm-1;  3 - Frequency, THz;  ( 4 - Attenuation, Db try to add later, this is not dB but something involving Db, need to check )
    switch(currentUnits)
    {
        case 0:
            return 10000000 / currentValue;
            break;
        case 1:
            return 10000 / currentValue;
            break;
        case 2:
            return currentValue;
            break;
        case 3:
            return currentValue * 3.3356409519815204957557671447492 * 1.0e-11 * 1.0e12;  // 0.01 is added in to take into account the conversion from centimeters to meters, the other multiplier 1.0e12 is to use THz
            break;
        default:
            return currentValue;
            break;
    }
}

// Minimum wavenumber
double Wavelength::getStartWn() { return this->startWn; }
void Wavelength::setStartWn(double newStartWn)
{
    this->startWn = newStartWn;
    findCalculatedResolAndNumPoints();  // Finding the calculated wavenumber resolution or number of data points
}
// Minimum wavelength in current units (setting and returning)
double Wavelength::getStartWlCurrUnits()
{
    // In case of wavelength the largest wavenumber corresponds to the smallest wavelength
    if (this->currentUnits < 2) { return fromDefaultToCurrent(endWn); }
    else { return fromDefaultToCurrent(startWn); }
}
void Wavelength::setStartWlCurrUnits(double newStartWlCurrUnits)
{
    if (this->currentUnits < 2) { endWn = fromCurrentToDefault(newStartWlCurrUnits); }
    else { startWn = fromCurrentToDefault(newStartWlCurrUnits); }
    findCalculatedResolAndNumPoints();  // Finding the calculated wavenumber resolution or number of data points
}

// Maximum wavenumber
double Wavelength::getEndWn() { return this->endWn; }
void Wavelength::setEndWn(double newEndWn)
{
    this->endWn = newEndWn;
    findCalculatedResolAndNumPoints();  // Finding the calculated wavenumber resolution or number of data points
}
// Minimum wavelength in current units (setting and returning)
double Wavelength::getEndWlCurrUnits()
{
    // In case of wavelength the largest wavenumber corresponds to the smallest wavelength
    if (this->currentUnits < 2) { return fromDefaultToCurrent(startWn); }
    else { return fromDefaultToCurrent(endWn); }
}
void Wavelength::setEndWlCurrUnits(double newEndWlCurrUnits)
{
    if (this->currentUnits < 2) { startWn = fromCurrentToDefault(newEndWlCurrUnits); }
    else { endWn = fromCurrentToDefault(newEndWlCurrUnits); }
    findCalculatedResolAndNumPoints();  // Finding the calculated wavenumber resolution or number of data points
}

// Method to get and set the number of data points
quint32 Wavelength::getNumPoints() { return this->numPoints; }
void Wavelength::setNumPoints(quint32 newNumPoints)
{
    // Saving the number of datapoints
    this->numPoints = newNumPoints;

    // Finding the calculated absolute wavenumber resolution and the number of data points
    findCalculatedResolAndNumPoints();
}
// Number of data points calculated from the absolute resolution and the current spectral range
quint32 Wavelength::getCalculatedNumPoints() { return this->calculatedNumPoints; }

// Wavenumber rsolustion mode
quint8 Wavelength::getWnResolMode() { return this->wnResolMode; }
void Wavelength::setWnResolMode(quint8 newWnResolMode)
{
    this->wnResolMode = newWnResolMode;
    findCalculatedResolAndNumPoints();  // Finding the calculated wavenumber resolution or number of data points
}

// Wavenumber resolution for calculations
double Wavelength::getWnResol() { return this->wnResol; }
void Wavelength::setWnResol(double newWnResol)
{
    this->wnResol = newWnResol;
    findCalculatedResolAndNumPoints();  // Finding the calculated wavenumber resolution or number of data points
}
// Wavenumber resolution for calculations calculated from the number of data points and the current calculation region
double Wavelength::getCalculatedWnResol() { return this->calculatedWnResol; }
// Wavenumber resolution for calculations returned in current units
double Wavelength::getWlResolCurrUnits()
{
    double midWn;  // Middle wavenumber, cm-1
    double maxWn;  // Wavenumber value shifted by the current resolution value
    double currUnitsWlOne;  // First wavelength in current units
    double currUnitsWlTwo;  // Second wavelength in current units
    double currUnitsResol;  // Resolution in current units


    if (currentUnits == 2)
    {
        return this->wnResol;
    }
    else  // In the conversion to other units the resolution will depend on the location of the spectral region specified
    {
        midWn = (startWn + endWn) * 0.5;
        maxWn = midWn + this->wnResol;
        currUnitsWlOne = fromDefaultToCurrent(midWn);
        currUnitsWlTwo = fromDefaultToCurrent(maxWn);
        currUnitsResol = fabs((double) (currUnitsWlOne - currUnitsWlTwo));
        return currUnitsResol;
    }
}
void Wavelength::setWlResolCurrUnits(double newWnResol)
{
    double midWn;  // Middle wavenumber, cm-1
    double midWlCurrUnits;  // Middle wavelength in current units
    double maxWlCurrUnits;  // Wavelength value shifted by the current resolution value
    double oneWn;  // First wavenumber value used to determine the resolution
    double twoWn;  // Second wavenumber value used to determine the resolution

    if (currentUnits == 2)
    {
        this->wnResol = newWnResol;
    }
    else  // In the conversion to other units the resolution will depend on the location of the spectral region specified
    {
        midWn = (startWn + endWn) * 0.5;
        midWlCurrUnits = fromDefaultToCurrent(midWn);
        maxWlCurrUnits = midWlCurrUnits + newWnResol;
        oneWn = fromCurrentToDefault(midWlCurrUnits);
        twoWn = fromCurrentToDefault(maxWlCurrUnits);
        this->wnResol = fabs((double) (oneWn - twoWn));
    }
    findCalculatedResolAndNumPoints();  // Finding the calculated wavenumber resolution or number of data points
}
// Calculated wavenumber resolution for calculations returned in current units
double Wavelength::getCalculatedWlResolCurrUnits()
{
    double midWn;  // Middle wavenumber, cm-1
    double maxWn;  // Wavenumber value shifted by the current resolution value
    double currUnitsWlOne;  // First wavelength in current units
    double currUnitsWlTwo;  // Second wavelength in current units
    double currUnitsResol;  // Resolution in current units


    if (currentUnits == 2)
    {
        return this->calculatedWnResol;
    }
    else  // In the conversion to other units the resolution will depend on the location of the spectral region specified
    {
        midWn = (startWn + endWn) * 0.5;
        maxWn = midWn + this->calculatedWnResol;
        currUnitsWlOne = fromDefaultToCurrent(midWn);
        currUnitsWlTwo = fromDefaultToCurrent(maxWn);
        currUnitsResol = fabs((double) (currUnitsWlOne - currUnitsWlTwo));
        return currUnitsResol;
    }
}

// Parameter indicating if the calculated resolution and the number of data points need to be calculated when other values (i.e. wavelength, resolution, number of data points, resolution mode)
bool Wavelength::getEnableRecalcOfCalculatedValues() { return this->enableRecalcOfCalculatedValues; }
void Wavelength::setEnableRecalcOfCalculatedValues(bool newEnableRecalcOfCalculatedValues) { this->enableRecalcOfCalculatedValues = newEnableRecalcOfCalculatedValues; }

// Distance from the line center in wavenumbers to run calculations up to
double Wavelength::getWnWing() { return this->wnWing; }
void Wavelength::setWnWing(double newWnWing) { this->wnWing = newWnWing; }

// Distance from the line center in half widths
double Wavelength::getWnWingHw() { return this->wnWingHw; }
void Wavelength::setWnWingHw(double newWnWingHw) { this->wnWingHw = newWnWingHw; }

// Variable indicating if the the spectral calcualtion range in the wings is expressed in wavenumbers or fractions of half widths
bool Wavelength::getWnWingAbs() { return this->wnWingAbs; }
void Wavelength::setWnWingAbs(bool newWnWingAbs) { this->wnWingAbs = newWnWingAbs; }

// Array with the main wavenumber grid values
QVector<double> *Wavelength::getWns() { return &(this->wns); }
double Wavelength::getWnsAt(int index) { return this->wns[index]; }

// Method to initialize the wavenumber grid based on the staring and ending wavenumbers as well as the wavenumber step
void Wavelength::initWavenumberGrid()
{
    int i;  // Loop variable
    quint32 auxNumWnPoints;  // Auxiliary nuber of data points variable
    double auxResol;         // Auxiliary wavenumber resolution value


    // Depending on the mode detemining the number of data points differently
    if (this->wnResolMode == 0)
    {
        auxNumWnPoints = ((this->endWn - this->startWn) / this->wnResol) + 1;
        auxResol = this->wnResol;
    }
    else if (this->wnResolMode == 1)
    {
        auxResol = (this->endWn - this->startWn) / (this->numPoints - 1);
        if (auxResol < this->wnResol)
        {
            auxNumWnPoints = ((this->endWn - this->startWn) / this->wnResol) + 1;
            auxResol = this->wnResol;
        }
        else
        {
            auxNumWnPoints = this->numPoints;
        }
    }
    else if (this->wnResolMode == 2)
    {
        auxNumWnPoints = this->numPoints;
        auxResol = (this->endWn - this->startWn) / (this->numPoints - 1);
    }

    // Determining the calculated absolute resolution and the number of data points values
    findCalculatedResolAndNumPoints();

    this->wns.resize(auxNumWnPoints);
    for (i = 0; i < auxNumWnPoints; i++)
    {
        this->wns[i] = this->startWn + auxResol * i;
    }
}

// Method to get the axis title for the currently selected units
QString Wavelength::getAxisTitle()
{
    return this->axisTitles[currentUnits];
}

// Finding the calcualated resolution and the number of data points
void Wavelength::findCalculatedResolAndNumPoints()
{
    double auxResol;  // Auxiliary reolustion variable


    // Recalculating the calculated values if the flag is set to do so
    if (this->enableRecalcOfCalculatedValues == true)
    {
        if (this->wnResolMode == 0)
        {
            this->calculatedWnResol = this->wnResol;
            this->calculatedNumPoints = ((this->endWn - this->startWn) / this->wnResol) + 1;
        }
        else if (this->wnResolMode == 1)
        {
            auxResol = (this->endWn - this->startWn) / (this->numPoints - 1);
            if (auxResol < this->wnResol)
            {
                this->calculatedWnResol = this->wnResol;
                this->calculatedNumPoints = ((this->endWn - this->startWn) / this->wnResol) + 1;
            }
            else
            {
                this->calculatedWnResol = (this->endWn - this->startWn) / (this->numPoints - 1);
                this->calculatedNumPoints = this->numPoints;
            }
        }
        else if (this->wnResolMode == 2)
        {
            this->calculatedWnResol = (this->endWn - this->startWn) / (this->numPoints - 1);
            this->calculatedNumPoints = this->numPoints;
        }
    }
}
