// bytran -|- spectral calculations for portable devices using the HITRAN database
//
// Developed by Dzianis Pliutau (Denis Plutov) and Kostya Roslyakov
// www.bytran.org | bytran@bytran.org
// Copyright (c) 2016 Dzianis Pliutau
// Distributed under the MIT license
//
// The MIT License (MIT)
// Copyright (c) 2016 Dzianis Pliutau
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation
// files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,
// modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software
// is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
// OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
// LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
// IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.



// Object is used to relay the wavelength information betweeen the C++ data objects and the QML interface
// All the data conversion code may better be moved into the Core module so the values retured are in the right
// units, otherwise this will have to be reimplemented for every component where units conversion is taking place
#ifndef WAVELENGTHQML_H
#define WAVELENGTHQML_H

#include <QObject>
#include <QLocale>
#include "bytrancore.h"

class WavelengthQml : public QObject
{
    Q_OBJECT

    Q_PROPERTY(QString startWavelength READ startWavelength WRITE setStartWavelength NOTIFY startWavelengthChanged)
    Q_PROPERTY(QString endWavelength READ endWavelength WRITE setEndWavelength NOTIFY endWavelengthChanged)
    Q_PROPERTY(QString resolWavelength READ resolWavelength WRITE setResolWavelength NOTIFY resolWavelengthChanged)
    Q_PROPERTY(quint8 units READ units WRITE setUnits NOTIFY unitsChanged)
    Q_PROPERTY(int numPoints READ numPoints WRITE setNumPoints NOTIFY numPointsChanged)
    Q_PROPERTY(quint8 wnResolMode READ wnResolMode WRITE setWnResolMode NOTIFY wnResolModeChanged)
    Q_PROPERTY(QString calculatedWlResolCurrUnits READ calculatedWlResolCurrUnits NOTIFY calculatedWlResolCurrUnitsChanged)
    Q_PROPERTY(quint32 calculatedNumPoints READ calculatedNumPoints NOTIFY calculatedNumPointsChanged)

public:

    explicit WavelengthQml(QObject *parent = nullptr, BytranCore *inCore = nullptr);
    WavelengthQml(const QString &startWavelength, const QString &endWavelength, const QString &resolWavelength, const int &units, QObject *parent=nullptr);

    // Start wavelength
    QString startWavelength() const;
    void setStartWavelength(const QString &startWavelength);

    // End wavelength
    QString endWavelength() const;
    void setEndWavelength(const QString &endWavelength);

    // Wavelength resolution
    QString resolWavelength() const;
    void setResolWavelength(const QString &resolWavelength);

    // Wavelength units
    quint8 units() const;
    void setUnits(const quint8 &units);

    // Number of points
    int numPoints() const;
    void setNumPoints(const int &numPoints);

    // Resolution specification mode
    quint8 wnResolMode() const;
    void setWnResolMode(const quint8 &wnResolMode);

    // Calculated resolution value
    QString calculatedWlResolCurrUnits() const;

    // Calculated number of wavelength data points
    quint32 calculatedNumPoints() const;

signals:

    void startWavelengthChanged();
    void endWavelengthChanged();
    void resolWavelengthChanged();
    void calculatedWlResolCurrUnitsChanged();
    void calculatedNumPointsChanged();
    void wnResolModeChanged();
    void unitsChanged();
    void numPointsChanged();

private slots:

    // Slot is used to update all values in the corresponding QML dialog
    void setUpdatedValues();

private:

    QString m_startWavelength;  // Starting wavelength in current wavelength units
    QString m_endWavelength;    // Ending wavelength in current wavelength units
    QString m_resolWavelength;  // Absoltute wavelength resolution in current wavelength units
    int m_numPoints;  // Number of data points
    int m_units;      // Wavelength units used in the interface
    quint8 m_wnResolMode;   // Wavelength resolution specification mode
    QString m_calculatedWlResolCurrUnits;  // Calculated wavelength resolution in currently selected units
    quint32 m_calculatedNumPoints;   // Calculated number of wavelength data points

    // Core object with the calculation logic etc.
    BytranCore *core;

    // Locale specific string to number and inverse conversions
    QLocale loc;

};

#endif // WAVELENGTHQML_H
