// bytran -|- spectral calculations for portable devices using the HITRAN database
//
// Developed by Dzianis Pliutau (Denis Plutov) and Kostya Roslyakov
// www.bytran.org | bytran@bytran.org
// Copyright (c) 2016 Dzianis Pliutau
// Distributed under the MIT license
//
// The MIT License (MIT)
// Copyright (c) 2016 Dzianis Pliutau
//
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation
// files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,
// modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software
// is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
// OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
// LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
// IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.



// Data object linking the vertical axis window with the core module performing the calculations
#ifndef YAXISQML_H
#define YAXISQML_H

#include <QObject>
#include <QLocale>
#include "bytrancore.h"


class YAxisQml : public QObject
{
    Q_OBJECT

    Q_PROPERTY(quint8 displayedValue READ displayedValue WRITE setDisplayedValue NOTIFY displayedValueChanged)
    Q_PROPERTY(QString minimum READ minimum WRITE setMinimum NOTIFY minimumChanged)
    Q_PROPERTY(QString maximum READ maximum WRITE setMaximum NOTIFY maximumChanged)
    Q_PROPERTY(quint8 units READ units WRITE setUnits NOTIFY unitsChanged)
    Q_PROPERTY(bool crossSectionAutoRecalc READ crossSectionAutoRecalc WRITE setCrossSectionAutoRecalc NOTIFY crossSectionAutoRecalcChanged)
    Q_PROPERTY(bool radianceAutoRecalc READ radianceAutoRecalc WRITE setRadianceAutoRecalc NOTIFY radianceAutoRecalcChanged)
    Q_PROPERTY(bool enableRulers READ enableRulers WRITE setEnableRulers NOTIFY enableRulersChanged)

public:

    Q_INVOKABLE void updateScaleValues();

    explicit YAxisQml(QObject *parent = 0, BytranCore *inCore = 0);
    YAxisQml(const QString &minimum, const QString &maximum, const int &units, QObject *parent=0);

    // Displayed units
    quint8 displayedValue() const;
    void setDisplayedValue(const quint8 &displayedValue);

    // Minimum
    QString minimum() const;
    void setMinimum(const QString &minimum);
    // Maximum
    QString maximum() const;
    void setMaximum(const QString &maximum);

    // Minimum absorption cross-section
    QString minimumCrossSection() const;
    void setMinimumCrossSection(const QString &minimumCrossSection);
    // Maximum absorption cross-section
    QString maximumCrossSection() const;
    void setMaximumCrossSection(const QString &maximumCrossSection);

    // Minimum radiance
    QString minimumRadiance() const;
    void setMinimumRadiance(const QString &minimumRadiance);
    // Maximum radiance
    QString maximumRadiance() const;
    void setMaximumRadiance(const QString &maximumRadiance);

    // Vertical scale units
    quint8 units() const;
    void setUnits(const quint8 &units);

    // Auto recalculation of the cross-sectional minimum and maximum values
    bool crossSectionAutoRecalc() const;
    void setCrossSectionAutoRecalc(const bool &crossSectionAutoRecalc);

    // Auto recalculation of the radiance minimum and maximum values
    bool radianceAutoRecalc() const;
    void setRadianceAutoRecalc(const bool &radianceAutoRecalc);

    // Methods to enable or disable the tracking rullers
    bool enableRulers() const;
    void setEnableRulers(const bool &enableRulers);

signals:

    void displayedValueChanged();
    void minimumChanged();
    void maximumChanged();
    void unitsChanged();

    // Auto recalculation of the cross-section and radiance min / max scale limits
    void crossSectionAutoRecalcChanged();
    void radianceAutoRecalcChanged();

    // Signals emitted when the ruler enabled state is changed
    void enableRulersChanged();

private slots:

    // Slot is used to update all values in the corresponding QML dialog
    void setUpdatedValues();

private:


    quint8 m_displayedValue;  // Displayed value (0 - Attenuation, 1 - Cross-section, 2 - Radiance)
    QString m_minimum;
    QString m_maximum;
    quint8 m_units;

    bool m_crossSectionAutoRecalc;  // Auto recalculation of the cross-section minimum and maximum scale values (true if enabled)
    bool m_radianceAutoRecalc;      // Auto recalculation of the radiance minimum and maximum scale values (true if enabled)

    // Enabled state of the tracking rulers
    bool m_enableRulers;

    // Core object with the calculation logic etc.
    BytranCore *core;

    // Locale specific string to number and inverse conversions
    QLocale loc;

};

#endif // YAXISQML_H
